from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
import numpy as np
from scipy.spatial.distance import cdist
from ase.data import atomic_numbers
import warnings
from matscipy.neighbours import neighbour_list as msp_nl


class ACSF(DescriptorBaseClass):

    name = 'ACSF'
    descriptor_type = 'local'

    def __init__(self, eta, rs, rc=5, xi=[], lamb=[], eta_ang=[], **kwargs):
        """
        Class for calculation of Behler Paranello features.

        Inputs:
           -- species: List of atomic numbers or atomic symbols
           -- eta: Exponent of radial functions
           -- rs: Offset of Gaussians. 
           -- rc: Cutoff radius
           -- xi: Polynomial exponents of angular functions
           -- lamb (da): Signs for angular functions
           -- eta_ang: Exponential exponents of angular functions.
        """        

        super().__init__(**kwargs)
        self.species = sorted([atomic_numbers[i] for i in self.environment.get_species()])
        
        self.eta = np.array(eta)
        self.rs = np.array(rs)
        self.xi = np.array(xi)
        self.lamb = np.array(lamb)
        self.eta_ang = np.array(eta_ang)
        self.rc = rc
        
        # Dimensions:
        if isinstance(eta, list):
            self.num_radial = len(eta)
        elif isinstance(rs, list):
            self.num_radial = len(rs)
        else:
            self.num_radial = 1
            
        self.num_angular = len(xi)

        self.dimension = (self.num_radial+self.num_angular) * len(self.species) + 1

    def create_features(self, atoms):
        """
        Calculates the Behler Paranello feature vectors for all atoms in a given structure.
        """
        # Matrix to save output:
        num_atoms = len(atoms)
        F = np.zeros((num_atoms, self.dimension))

        # Calculate distances:
        i1, i2, dists = msp_nl('ijd', atoms, self.rc)
        dists = dists.reshape(-1, 1)
        # Radial functions given by:
        for i in range(num_atoms):
            args = i1 == i
            relevant_dists = dists[args].reshape(-1, 1)
            for c, species in enumerate(self.species):
                filter = atoms.numbers[i2[args]] == species
                d = relevant_dists[filter]
                F[i, c*self.num_radial:(c+1)*self.num_radial] = np.sum(np.exp(-self.eta*(d-self.rs)**2)*self.cutoff(d), axis = 0)

        F[:, -1] = atoms.numbers
        return F

        for i in range(num_atoms):
            neigh_mask = i1==i
            for Rij, rij, j in zip(D[neigh_mask], dists[neigh_mask].ravel(), i2[neigh_mask]):
                for Rik, rik, k in zip(D[neigh_mask], dists[neigh_mask].ravel(), i2[neigh_mask]):
                    #print(i, j, k)
                    #if j != k and k != i: Equal conditions?
                    if j < k:
                        Rjk = Rik-Rij
                        rjk = np.sqrt(Rjk@Rjk)
                        theta = (Rij@Rik)/(np.sqrt(Rij@Rij)*np.sqrt(Rik@Rik))
                        F[i, self.num_radial::] += ((1-self.lamb*theta)**self.xi
                        *np.exp(-self.eta_ang*(rij**2+rik**2+rjk**2)/rc2)
                        *self.cutoff(rij)*self.cutoff(rik)*self.cutoff(rjk))
        F[:, self.num_radial::] *= 2**(1-self.xi)

        return F

    def calculate_specifics(self, atoms, index):
        """
        Arguments:
        -- Atoms: Atoms object
        -- Index: List/np.array of indicies of the atoms to calculate features for. 
        """

        if type(index) == int or type(index)==np.int64:
            index = [index]
        
        F = np.zeros((len(index), self.dimension), dtype=float)
        
        num_atoms = len(atoms)
        for ii, idx in enumerate(index):
            allowed_atoms = [i for i in range(num_atoms) if i != idx]
            
            d = cdist(atoms.positions, atoms.positions[idx].reshape(1, -1))
            F[ii, 0:self.num_radial] = np.sum((np.exp(-self.eta*(d[allowed_atoms]-self.rs)**2
                                        /self.rc**2)*self.cutoff(d[allowed_atoms])), axis=0)

            for j in range(num_atoms):
                if j == idx:
                    continue
                Rij = atoms.positions[j, :]-atoms.positions[idx, :]
                rij = d[j]
                Fij = self.cutoff(rij)
                for k in range(j+1, num_atoms):
                    if k == idx:
                        continue
                    Rjk = atoms.positions[k, :]-atoms.positions[j, :]
                    Rik = atoms.positions[k, :]-atoms.positions[idx, :]
                    theta = (Rij@Rik)/(np.sqrt(Rij@Rij)*np.sqrt(Rik@Rik))
                    rjk = np.sqrt(Rjk@Rjk)
                    rik = d[k]
                    Fik = self.cutoff(rik); Fjk = self.cutoff(rjk)
                    F[ii, self.num_radial::] += ((1-self.lamb*theta)**self.xi
                    *np.exp(-self.eta_ang*(rij**2+rik**2+rjk**2)/self.rc**2)
                                                  *Fij*Fjk*Fik)

            F[:, self.num_radial::] *= 2**(1-self.xi)   
        return F

    def calculate_jacobian(self, atoms):
        """
        Calculate the Jacobian of the feature vector at the current positions
        """

        N = len(atoms)
        
        # Arrays for saving:
        J = np.zeros((self.dimension*N, 3*N))
        
        # Calculate distances:
        rij = cdist(atoms.positions, atoms.positions)
        X = atoms.positions

        for i in range(N):
            for j in range(N):
                if rij[i, j] < self.rc:
                    for x in range(3):
                        if i != j:
                            J[i*self.dimension:(i+1)*self.dimension, j*3+x] = (
                            np.exp(-self.eta*(rij[i, j]-self.rs)**2/self.rc**2)
                            * (-2*self.eta*(rij[i, j]-self.rs)/self.rc**2 * self.cutoff(rij[i, j])
                            + self.cutoff_derivative(rij[i, j])) * (X[i, x] - X[j, x])/rij[i, j])

                            J[i*self.dimension:(i+1)*self.dimension, i*3+x] = (
                            J[i*self.dimension:(i+1)*self.dimension, j*3+x])
            
        return J                                        
                            
    def cutoff(self, r):
        return (r<=self.rc)*0.5*(1+np.cos(np.pi*r/self.rc))

    def cutoff_derivative(self, r):
        return -0.5*np.pi/self.rc * np.sin(np.pi*r/self.rc)

    @classmethod
    def from_species(cls, species, **kwargs):
        from ase import Atoms
        from agox.environments import Environment
        environment = Environment(template=Atoms(''), symbols=''.join(species), use_box_constraint=False,
                                  print_report=False)
        return cls(environment=environment, **kwargs)

    def get_number_of_centers(self, atoms):
        return len(atoms)



import pytest
from agox.test.test_utils import get_test_data, test_data_dicts
from ase.io import read
from agox.utils.filters.feature_distance import FeatureDistanceFilter
from agox.models.descriptors.spectral_graph_descriptor import SpectralGraphDescriptor

@pytest.mark.ray
@pytest.mark.parametrize('test_data_dict', test_data_dicts)
def test_feature_distance_filter(test_data_dict):
    path = test_data_dict['path']

    # Load structures from directory
    all_structures = read(path, ':') 

    # Setup a descriptor
    descriptor = SpectralGraphDescriptor.from_atoms(all_structures[0])

    # Setup a filter, structures closer than threshold are considered duplicates
    threshold = 1e-8
    feature_filter = FeatureDistanceFilter(descriptor, threshold)

    # Filter the structures
    filtered_structures, filtered_indices = feature_filter.filter(all_structures)

    print('Before filtering: ', len(all_structures))
    print('After filtering: ', len(filtered_structures))


"""
Evaluators are responsible for evaluating the objective function of a candidate.
The main two flavours of evaluators are single-point and local optimization.

Single-point evaluators evaluate the objective function at a single point, 
whereas local optimization evaluators perform a local optimization starting from 
the candidate's coordinates.
"""
from .single_point import SinglePointEvaluator
from .local_optimization import LocalOptimizationEvaluator
from .rattle import RattleEvaluator

__all__ = [
    'SinglePointEvaluator',
    'LocalOptimizationEvaluator',
    'RattleEvaluator']


import numpy as np
from timeit import default_timer as dt
from agox.observer import Observer, ObserverHandler
from agox.writer import Writer, agox_writer
from agox.module import Module
from agox.databases.ABC_database import DatabaseBaseClass

class Tracker(Observer, Writer):
    """
    Logger that looks at modules and logs properties they have assigned each iteration. 
    Also is responsible for timings. 
    """

    name = 'Tracker'

    def __init__(self, agox, save_name=None):
        order = agox.get_max_order() + 2
        all_gets, all_sets = agox.get_set_match()
        gets = {key:key for key in all_gets}
        Observer.__init__(self, order=order, gets=gets)
        Writer.__init__(self, verbose=True, use_counter=False, prefix='')

        self.tracked_attributes = {}
        self.state_loggers = []

        self.timer = Timer()
        self.timer.attach(agox)

        self.find_modules(agox)
        self.find_trackable_attributes()
        self.add_observer_method(self.update, sets={}, gets=gets, order=order)
        self.attach(agox)

        # Naming for files:
        database_file_name = self.find_database_name()
        if save_name is None:
            self.save_name = database_file_name + '_tracker'
        else:
            self.save_name = save_name

        agox.attach_finalization('tracker_save', self.save)

    def find_modules(self, agox):
        modules = [obs.class_reference for obs in agox.observers.values()]
        outer_modules = {element.cache_key:element for element in modules}

        all_modules = {key:val for key, val in outer_modules.items()}

        for module in outer_modules.values():
            submodules = module.find_submodules(only_dynamic=False)

            for submodule in submodules.values():
                all_modules[submodule.cache_key] = submodule

        self.modules_to_track = all_modules

    def find_database_name(self):
        database_filename = 'tracker'            
        for module in self.modules_to_track.values():
            if issubclass(module.__class__, DatabaseBaseClass):
                if hasattr(module, 'filename'):
                    database_filename = module.filename
                    # Without extension:
                    database_filename = database_filename.split('.')[0]
                    break
        return database_filename
            
    def get_attribute_name(self, module, attribute_key):
        return f'{module.__name__}.{attribute_key}'

    def find_trackable_attributes(self):
        for key, module in self.modules_to_track.items():
            for attribute_key in module.tracked_attributes:
                attribute_name = self.get_attribute_name(module, attribute_key)
                self.tracked_attributes[attribute_name] = []

        self.tracked_attributes['Timer.tracker.update'] = []


    @agox_writer
    @Observer.observer_method
    def update(self, state):
        t0 = dt()

        for module in self.modules_to_track.values():
            attribute_dict = module.get_tracked_attributes()
            for keys, tracked_attribute in attribute_dict.items():
                attribute_name = self.get_attribute_name(module, keys)
                self.tracked_attributes[attribute_name].append(tracked_attribute)

        for state_logger_tuple in self.state_loggers:
            function, name, gets, tracking_name = state_logger_tuple
            self.tracked_attributes[tracking_name].append(function(state.get_from_cache(self, gets)))

        t1 = dt()
        self.tracked_attributes['Timer.tracker.update'].append(t1-t0)

        
    def add_state_tracker(self, function, name, gets):
        tracking_name = f'state.{gets}.{name}'
        self.state_loggers.append((function, name, gets, tracking_name))
        self.tracked_attributes[tracking_name] = []

    def save(self):       
        np.savez(self.save_name+'.npz', **self.tracked_attributes)

    @staticmethod
    def load(filename):
        return {key:value for key, value in np.load(filename).items()}

class Timer(Observer, Writer):

    name = 'Timer'

    def __init__(self, save_frequency=100, **kwargs):
        """
        Logger instantiation. 

        The 'Logger' is the observer that is attached to main and prints 
        the log report. 

        Parameters:
        save_frequency: int
            How often to save to save the log to disk.
        """

        Observer.__init__(self, **kwargs)
        Writer.__init__(self, verbose=True, use_counter=False)
        self.timers = []
        self.ordered_keys = []
        self.save_frequency = save_frequency

        self.add_tracked_attribute('iteration_time')

    def attach(self, main):
        """
        Attaches to main. 

        Three things happen:
        1.  LogEntry's are created for each observer, which may also recursively 
            create LogEntry's for observers of observers and so on.
        2.  The 'Logger' attaches an 'report_logs' as an Observer to main, so 
            that the log is printed. 
        3.  A finalization method is added so that the log is saved when AGOX 
            finishes. 

        Parameters
        -----------
        main: AGOX object.
            The main AGOX class from agox/main.py
        """
        # Want to get all other observers: 
        observers = main.observers

        # Understand their order of execution:
        keys = []; orders = []
        for observer_method in observers.values():
            keys.append(observer_method.key) 
            orders.append(observer_method.order)
        sort_index = np.argsort(orders)
        sorted_keys = [keys[index] for index in sort_index]
        sorted_observers = [observers[key] for key in sorted_keys]

        # Attach log obserers to time them:
        for observer_method, key in zip(sorted_observers, sorted_keys):
            observer_timer = TimerEntry(observer_method)
            observer_timer.attach(main)
            self.timers.append(observer_timer)
    
        # Also attach a reporting:
        self.add_observer_method(self.report_logs, sets={}, gets={}, order=np.max(orders)+1)
        
        # Attach final dumping of logs:
        #main.attach_finalization('Logger.dump', self.log.save_logs)

        super().attach(main)

    @agox_writer
    @Observer.observer_method
    def report_logs(self, state):
        """
        Calls the 'log_report' of the log object.

        Saves the log to disk depending on self.save_frequency. 
        """
        total_time = np.sum([entry.get_current_timing() for entry in self.timers])
        self.iteration_time = total_time
        self.writer('Total time {:05.2f} s '.format(total_time))

        for entry in self.timers:
            report = entry.get_iteration_report()
            for line in report:
                self.writer(line)
        
class TimerEntry(Observer, Writer, Module):

    name = 'TimerEntry'

    def __init__(self, observer_method):
        """
        TimerEntry class. 
        
        Given an ObserverMethod the LogEntry attaches observers around that method 
        in the observer-loop to time the ObserverMethod. 

        If the Observer that ObserverMethod comes from is an instance of an 
        ObserverHandler it rescursively attaches other LogEntry instances to 
        the observers of that ObserverHandler to time those too. 

        Parameters
        -----------
        observer_method: ObserverMethod object. 
            An instance of a ObserverMethod (agox/observer.py) to attach around. 
        """

        Observer.__init__(self, order=observer_method.order)
        Writer.__init__(self, verbose=True, use_counter=False)
        self.timings = []
        self.name = 'Timer.' + observer_method.name
        self.observer_name = observer_method.name

        # Time sub-entries:
        self.sub_entries = {}
        self.recursive_attach(observer_method)

        self.add_tracked_attribute('timing')

    def attach(self, main):
        """
        Attachs class methods to the observer loop of main. 

        Parameters
        -----------
        main: AGOX object.
            The main AGOX class from agox/main.py
        """

        self.add_observer_method(self.start_timer, sets=self.sets[0], gets=self.gets[0], order=self.order[0]-0.01)
        self.add_observer_method(self.end_timer, sets=self.sets[0], gets=self.gets[0], order=self.order[0]+0.01)
        super().attach(main)

    def start_timer(self, state, *args, **kwargs):
        """
        Method attached as an observer to start the timing. 
        """
        self.timings.append(-dt())
        
    def end_timer(self, state, *args, **kwargs):
        """
        Method attached as an observer to end the timing.
        """
        if len(self.timings):
            self.timings[-1] += dt()

        self.timing = self.timings[-1]

    def get_current_timing(self):
        """
        Get most recent timing. 

        Returns
        --------
        float
            Timing of most recent iteration.
        """

        if len(self.timings):
            return self.timings[-1]
        else:
            # print(f'{self.name}: Somehow the timer failed - Havent figure out why this happens sometimes - {len(self.timings)}')
            return 0

    def get_sub_timings(self):
        """
        Get timing of sub entries, relevant when the object self is attached around 
        is also an ObserverHandler that has other Observers, such as a Database 
        that may have a model listening to it. 

        Returns
        --------
        List or float
            If sub_entries are present then returns a list of those subtimings 
            otherwise returns a throguh 'get_current_timing'.
        """
        if len(self.sub_entries):
            sub_timings = []
            for sub_entry in self.sub_entries.values():
                sub_timings.append(sub_entry.get_sub_timings())
            return sub_timings
        else:
            return self.get_current_timing()

    def recursive_attach(self, observer_method):
        """
        Attach recursively when the class that ObserverMethod originates from 
        is an instance of ObserverHandler.

        Parameters
        -----------
        observer_method: ObserverMethod object. 
            An instance of a ObserverMethod (agox/observer.py) self is attached 
            around. 
        """
        if issubclass(observer_method.class_reference.__class__, ObserverHandler):

            if not observer_method.name == observer_method.class_reference.dispatch_method:
                return
            
            # Want to get all other observers: 
            class_reference = observer_method.class_reference # The class the inbound observer method comes from.
            observer_dicts = observer_method.class_reference.observers # Dict of observer methods.


            # May sometimes find objects that already have LogEntry observers 
            # attached, so we remove those. 
            methods_to_delete = []
            for key, observer_method in observer_dicts.items():
                if 'LogEntry' in observer_method.name:
                    methods_to_delete.append(observer_method)

            for method in methods_to_delete:
                class_reference.delete_observer(method)                    

            # Understand their order of execution:
            keys = []; orders = []
            for key, observer_dict in observer_dicts.items():
                keys.append(key) 
                orders.append(observer_dict['order'])
            sort_index = np.argsort(orders)
            sorted_keys = [keys[index] for index in sort_index]
            sorted_observer_dicts = [observer_dicts[key] for key in sorted_keys]

            # Attach log entry observers to time them:
            for observer_method, key in zip(sorted_observer_dicts, sorted_keys):
                self.add_sub_entry(observer_method)
                self.sub_entries[observer_method.key].attach(class_reference)

    def add_sub_entry(self, observer_method):
        """
        Add a sub entry which is an other instance of LogEntry. 

        Parameters
        -----------
        observer_method: ObserverMethod object. 
            An instance of a ObserverMethod (agox/observer.py) from an Observer 
            that is observing the Observer self is attached around. 
            E.g. if self is attached around a Database then a Model may be 
            observing that database with an ObserverMethod.  
        """
        self.sub_entries[observer_method.key] = TimerEntry(observer_method)

    def get_iteration_report(self, report=None, offset=''):
        """
        Recursively generate iteration report. 

        Parameter
        ----------
        report: None or a list. 
            If None an empty list is created. Report strings are appended to that
            list.
        offset: str, 
            Indentation, reports from subentries are offset by 4 spaces. 

        Returns
        --------
        list
            The 'report' list of strings is returned. 
        """
        if report is None:
            report = [] # List of strings:

        report.append(offset + ' {} - {:05.2f} s'.format(self.observer_name, self.get_current_timing()))

        for sub_entry in self.sub_entries.values():
            report = sub_entry.get_iteration_report(report=report, offset=offset+' ' * 4)

        return report


from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

import numpy as np
from ase import Atoms


class SparsifierBaseClass(ABC):
    """
    Abstract class for sparsifiers.

    A sparsifier is a function that takes a matrix X with rows corresponding to feaures.
    It returns a matrix with the same number of columns, possibly together with which rows are selected.


    Parameters
    ----------
    m_points : int
        Number of points to select

    Methods
    -------
    sparsify(atoms: List[Atoms] X): np.ndarray) -> np.ndarray:
        Sparsify the data
    """

    def __init__(
        self,
        m_points: int = 1000,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        descriptor : DescriptorBaseClass
            Descriptor to use for computing the features
        m_points : int
            Number of points to select
        descriptor_type : str
            Type of descriptor to use. Can be "global" or "local"

        """
        super().__init__(**kwargs)
        self.m_points = m_points

    @abstractmethod
    def sparsify(self, X: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns a matrix with the same number of columns together with which rows are selected.

        Parameters
        ----------
        X : np.ndarray
            Matrix with rows corresponding to features

        Returns
        -------
        X_sparsified : array-like, shape (m_points, n_samples)
            Matrix with rows corresponding to feaures.
        selected : array-like, shape (m_points,)

        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        return NotImplementedError

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return self.sparsify(X)

    def __add__(self, other: "SparsifierBaseClass"):
        return SumSparsifier(s0=self, s1=other)


class SumSparsifier(SparsifierBaseClass):
    """
    Sum of two sparsifiers

    Parameters
    ----------
    s0 : SparsifierBaseClass
        First sparsifier
    s1 : SparsifierBaseClass
        Second sparsifier

    """

    def __init__(
        self,
        s0: SparsifierBaseClass,
        s1: SparsifierBaseClass,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.s0 = s0
        self.s1 = s1

        self.m_points = s1.m_points

    def sparsify(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns a matrix with the same number of columns together with which rows are selected.

        Parameters
        ----------
        X : np.ndarray
            Matrix with rows corresponding to features

        Returns
        -------
        X_sparsified : array-like, shape (m_points, n_features)
            Matrix with rows corresponding to feaures.
        selected : array-like, shape (m_points,)
        """
        # select with s0
        Xm0, idxs0 = self.s0(X=X)
        Xm, idxs1 = self.s1(X=Xm0)
        return Xm, idxs0[idxs1]

    @property
    def name(self):
        return f"{self.s0.name}+{self.s1.name}"



import re
from pathlib import Path

def convert_to_seconds(time_str):
    h, m, s = time_str.split(':')
    return int(h) * 3600 + int(m) * 60 + int(s)

class LogReader():

    def __init__(self, log_file: Path):
        self.log_file = log_file.resolve()
        if self.log_file.exists():
            self.events, self.slurm_env = self.analyze_log(log_file)
            self.log_read = True
        else: 
            self.log_read = False

    def find_timestamp(self, line: str):
        try:
            time = re.findall('\d{2}:\d{2}:\d{2}', line)[0]
        except:
            time = False
        return time

    def analyze_log(self, log_file: Path):

        slurm_env = {}
        events = {}

        with open(log_file, 'r') as f:
            lines = f.readlines()

            for line in lines:
                time = self.find_timestamp(line)

                if time is False:
                    continue
                
                try:
                    split_line = line.split(' - ')
                    output = split_line[1]
                except: 
                    continue

                if 'SLURM' in output:
                    
                    split_line = output.split(':')
                    key, value = split_line[0], split_line[1]
                    key = key.strip()
                    value = value.strip()
                    if len(value) > 0:
                        slurm_env[key] = value


                if 'Event' in output:

                    event_index = int(re.findall('\d', output)[0])
                    events[event_index] = (time, output.strip())

            return events, slurm_env

    def get_out_path(self):
        """
        Not super robust.
        """
        directory = self.slurm_env['SLURM_SUBMIT_DIR']
        array_id = self.slurm_env['SLURM_ARRAY_TASK_ID']
        job_id = self.slurm_env['SLURM_ARRAY_JOB_ID']

        output_path=Path(directory, f"{array_id}/slurm-{job_id}-{array_id}.out")
        return output_path
    
    def print_output(self):
        path = self.get_out_path()

        if path.exists():
            with open(path, 'r') as f:
                lines = f.readlines()
                for line in lines:
                    print(line.strip())
        
        else:
            print(f"Output file {path} not found.")

    def get_slurm_env_var(self, key):
        return self.slurm_env.get(key, None)

    def get_partition(self):
        return self.get_slurm_env_var('SLURM_JOB_PARTITION')
    
    def get_node(self):
        return self.get_slurm_env_var('SLURM_JOB_NODELIST')

    def get_state(self):
        return self.events.get(6, None) is not None
    
    def get_most_recent_event_index(self):
        return max(self.events.keys())
    
    def get_execution_time(self):                
        start_time = convert_to_seconds(self.events[0][0])
        end_time = convert_to_seconds(self.events[self.get_most_recent_event_index()][0])
        return end_time - start_time


import pytest

def pytest_addoption(parser):
    parser.addoption('--rtol', type=float, default=1e-05)
    parser.addoption('--atol', type=float, default=1e-08)
    parser.addoption('--create_mode', default=False, action='store_true')

@pytest.fixture
def cmd_options(request):
    return {'tolerance':{'rtol':request.config.getoption('rtol'), 'atol':request.config.getoption('atol')},
            'create_mode':request.config.getoption('create_mode')}


@pytest.fixture(scope="module", autouse=True)
def ray_fix():
    import ray 
    from agox.utils.ray import reset_ray_pool
    if ray.is_initialized():
        reset_ray_pool()
    yield

def pytest_sessionstart(session):
    from agox.utils.ray import ray_startup
    ray_startup(cpu_count=2, memory=None, tmp_dir=None, include_dashboard=False, max_grace_period=0.0)



"""
Generators are are responsible for generating new configurations. 
Typically there are two types of generators: 

* Those that can generate a candidate by modifying one or more configurations. 
* Those that can generate a candidate without any input configurations.

The first catagory of generators require a Sampler to sample the configurations
that it uses as a basis for generating a new candidate. The second category of
generators do not require a sampler.    

The generators have to generate a candidate that follows the rules set by the
environment, that is

* The correct stoichoimetry
* Inside the confinement box. 
"""
from agox.generators.ABC_generator import GeneratorBaseClass
from agox.generators.random import RandomGenerator
from agox.generators.rattle import RattleGenerator
from agox.generators.replace import ReplaceGenerator
from agox.generators.reuse import ReuseGenerator
from agox.generators.permutation import PermutationGenerator
from agox.generators.steepest_descent import SteepestDescentGenerator
from agox.generators.cog import CenterOfGeometryGenerator
from agox.generators.sampling import SamplingGenerator
from agox.generators.ce_generator import ComplementaryEnergyGenerator
from agox.generators.MD import MDgenerator

from agox.generators.symmetry import SymmetryGenerator
from agox.generators.symmetry_rattle import SymmetryRattleGenerator
from agox.generators.symmetry_permutation import SymmetryPermutationGenerator

__all__ = ["GeneratorBaseClass", 
           "RandomGenerator", 
           "RattleGenerator", 
           "ReplaceGenerator",
           "ReuseGenerator", 
           "PermutationGenerator",
           "SteepestDescentGenerator", 
           "CenterOfGeometryGenerator", 
           "SamplingGenerator", 
           "ComplementaryEnergyGenerator", 
           "SymmetryGenerator", 
           "SymmetryRattleGenerator", 
           "SymmetryPermutationGenerator", 
           "MDGenerator"]

__all__ = []


import os
from time import sleep
from agox.databases.database import Database
from .database_utilities import *

from agox.observer import Observer
from agox.writer import agox_writer

class ConcurrentDatabase(Database):

    """
    A database that can be used to store information from multiple concurrent instances of AGOX.
    """

    init_statements = ["""create table structures (
    id integer primary key autoincrement,
    ctime real,
    positions blob,
    energy real,
    type blob,
    cell blob,
    forces blob, 
    pbc blob,
    template_indices blob,
    iteration int,
    worker_number int
    )""", 
    
    """CREATE TABLE text_key_values (
    key TEXT,
    value TEXT,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",

    """CREATE TABLE float_key_values (
    key TEXT,
    value REAL,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""", 

    """CREATE TABLE int_key_values (
    key TEXT,
    value INTEGER,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""", 
    
    """CREATE TABLE boolean_key_values (
    key TEXT,
    value INTEGER,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""", 
    
    """CREATE TABLE other_key_values (
    key TEXT,
    value BLOB,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))"""]

    # Pack: Positions, energy, type, cell, forces, pbc, template_indices, iteration, worker_number
    # Unpack: ID, time, -//-
    pack_functions = [blob, nothing, blob, blob, blob, blob, blob, nothing, nothing]
    unpack_functions = [nothing, nothing, deblob, nothing, deblob, deblob, deblob, deblob, deblob, nothing, nothing]

    def __init__(self, worker_number=0, total_workers=1, sleep_timing=1, 
        sync_frequency=50, sync_order=None, synchronous=True, **kwargs):
        super().__init__(**kwargs)

        self.storage_keys.append('worker_number')
        self.worker_number = worker_number
        self.total_workers = total_workers
        self.sync_frequency = sync_frequency
        self.synchronous = synchronous
        self.sleep_timing = sleep_timing

        self.filename_ready = self.filename[:-3] + '_WORKER{}_READY{}'                
        self.filename_done = self.filename[:-3] + '_WORKER{}_DONE{}'

        if sync_order is None:
            self.sync_order = self.order[0] + 0.1
        else:
            self.sync_order = sync_order

        #self.add_observer_method(self.sync_database, gets={}, sets={}, order=self.sync_order)

        self._initialize()

    def _init_storage(self):
        super()._init_storage()

    def store_information(self, candidate):
        super().store_information(candidate=candidate)
        self.storage_dict['worker_number'].append(self.worker_number)

    def store_candidate(self, candidate, accepted=True, write=True, dispatch=True):
        candidate.add_meta_information('worker_number', self.worker_number)
        candidate.add_meta_information('iteration', self.get_iteration_counter())
        super().store_candidate(candidate, accepted=accepted, write=write, dispatch=dispatch)

    def db_to_candidate(self, structure, meta_dict=None):
        candidate = super().db_to_candidate(structure, meta_dict=meta_dict)
        candidate.add_meta_information('worker_number', structure['worker_number'])
        candidate.add_meta_information('iteration', structure['iteration'])
        candidate.add_meta_information('id', structure['id'])

        return candidate

    @agox_writer
    @Observer.observer_method
    def store_in_database(self, state):
        
        evaluated_candidates = state.get_from_cache(self, self.get_key)
        anything_accepted = False
        for j, candidate in enumerate(evaluated_candidates):

            # Dispatch to observers only when adding the last candidate. 
            dispatch = (j+1) == len(evaluated_candidates)

            if candidate: 
                self.writer('Energy {:06d}: {}'.format(self.get_iteration_counter(), candidate.get_potential_energy()), flush=True)
                self.store_candidate(candidate, accepted=True, write=True, dispatch=False)
                anything_accepted = True

            elif candidate is None:
                dummy_candidate = self.candidate_instanstiator(template=Atoms())
                dummy_candidate.set_calculator(SinglePointCalculator(dummy_candidate, energy=float('nan')))

                # This will dispatch to observers if valid data has been added but the last candidate is None. 
                self.store_candidate(candidate, accepted=False, write=True, dispatch=False)

        # This is the difference compared to the base-class method.
        self.sync_database()

        if anything_accepted:
            self.dispatch_to_observers(self, state)


    # @agox_writer
    # @Observer.observer_method
    def sync_database(self):
        if self.decide_to_sync():            
            if self.synchronous:
                self.synchronous_update()
            else:
                self.asynchronous_update()

    def synchronous_update(self):
        # write file
        iteration = self.get_iteration_counter()
        with open(self.filename_ready.format(self.worker_number, iteration), mode='w'): pass
        
        self.writer('Attempting to sync database')
        # Make sure the database contains all the expected information from all workers.
        state = self.check_status()
        print_string  = ''
        while not state:
            sleep(self.sleep_timing)
            state = self.check_status()
            print_string += '.'
        if len(print_string) > 0:
            self.writer(print_string)
        
        # Restore the database to memory. 
        # This will change the order of candidates in the Database, so be careful if another module relies on that!
        self.restore_to_memory()
        self.writer('Succesfully synced database')

        # write success file
        with open(self.filename_done.format(self.worker_number, iteration), mode='w'): pass

        if self.worker_number == 0: # i.e. i'm the master!
            state = self.cleanup()                    
            while not state:
                sleep(self.sleep_timing)
                state = self.cleanup()                    

        self.writer('Number of candidates synced from database {}'.format(len(self)))

    def asynchronous_update(self):
        self.writer('Before asynchronous update: {}'.format(len(self)))
        self.restore_to_memory()
        self.writer('After asynchronous update: {}'.format(len(self)))

    def check_status(self):
        expected_iteration = self.get_iteration_counter()
        state = True
        for i in range(self.total_workers):
            if not os.path.exists(self.filename_ready.format(i, expected_iteration)):
                state = False
        
        return state

    def cleanup(self):
        expected_iteration = self.get_iteration_counter()
        state = True
        for i in range(self.total_workers):
            if not os.path.exists(self.filename_done.format(i,expected_iteration)):
                state = False
        if state:
            for i in range(self.total_workers):
                os.remove(self.filename_ready.format(i,expected_iteration))
                os.remove(self.filename_done.format(i,expected_iteration))
        return state

    def decide_to_sync(self):
        return self.get_iteration_counter() % self.sync_frequency == 0

    def get_all_candidates(self, respect_worker_number=False):
        if respect_worker_number:
            all_candidates = []
            for candidate in self.candidates:
                if candidate.get_meta_information('worker_number') == self.worker_number:
                    all_candidates.append(candidate)
            return all_candidates
        else:
            return super().get_all_candidates()




import functools
from typing import List
import numpy as np

from agox import __version__

ICON = """
       _            _  _  _        _  _  _  _    _           _ 
     _(_)_       _ (_)(_)(_) _   _(_)(_)(_)(_)_ (_)_       _(_)
   _(_) (_)_    (_)         (_) (_)          (_)  (_)_   _(_)  
 _(_)     (_)_  (_)    _  _  _  (_)          (_)    (_)_(_)    
(_) _  _  _ (_) (_)   (_)(_)(_) (_)          (_)     _(_)_     
(_)(_)(_)(_)(_) (_)         (_) (_)          (_)   _(_) (_)_   
(_)         (_) (_) _  _  _ (_) (_)_  _  _  _(_) _(_)     (_)_ 
(_)         (_)    (_)(_)(_)(_)   (_)(_)(_)(_)  (_)         (_)  v{}_{} \n
"""

def get_git_revision_short_hash() -> str:
    import subprocess
    import os
    import agox

    try:
        dir_path = os.path.dirname(agox.__path__[0])
        version_string = subprocess.check_output(['git', f'--git-dir={dir_path}/.git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()
    except subprocess.CalledProcessError:
        version_string = 'unknown'
    except BlockingIOError:
        version_string = 'unknown'
    return version_string

LINE_LENGTH = 79
PADDING_CHARACTER = '='
TERMINATE_CHARACTER = '|'

def get_icon():
    version_string = get_git_revision_short_hash()
    return ICON.format(__version__, version_string)

def line_breaker(string: str, tab_size: int = 4) -> List[str]:
    """Break a long string into lines to be used in `pretty_print`.

    Parameters
    ----------
    string : str
        The long string to break into lines
    tab_size : int, optional
        The width of a tab character in number of space characters,
        by default 4

    Returns
    -------
    List[str]
        The original string split into individual lines
    """

    string = string.replace('\t', ' ' * tab_size)

    text_length = LINE_LENGTH - 4  # lines are padded with 4 characters
    lines = []
    remainder = string  # every iteration, remainder will be shortened

    while len(remainder) > text_length:
        max_line = remainder[:text_length + 1]  # there might be a space at the end
        space_index = max_line.rfind(' ')
        if space_index != -1:  # space found
            lines.append(remainder[:space_index])
            remainder = remainder[space_index + 1:]  # remove the space
        else:  # no space found, simply break the line at the maximum length
            lines.append(remainder[:text_length])
            remainder = remainder[text_length:]

    lines.append(remainder)  # final piece of the string
    return lines
        
def header_print(string):
    
    string = ' ' + string + ' '
    num_markers = int((LINE_LENGTH - len(string))/2) - 1
    header_string = TERMINATE_CHARACTER + num_markers * PADDING_CHARACTER + string + num_markers * PADDING_CHARACTER + TERMINATE_CHARACTER

    if len(header_string) < LINE_LENGTH:
        header_string = header_string[:-2] + PADDING_CHARACTER*2 + TERMINATE_CHARACTER        

    print(header_string, flush=True)

def pretty_print(string, *args, **kwargs):
    string = str(string)
    for arg in args:
        string += str(arg)

    all_strings = line_breaker(string)
    for string in all_strings:
        string = TERMINATE_CHARACTER + ' ' + string + (LINE_LENGTH - len(string)-3) * ' ' + TERMINATE_CHARACTER       
        print(string, **kwargs)

def agox_writer(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.use_counter:
            header_print(self.name)
        func(self, *args, **kwargs)
        if self.use_counter and len(self.lines_to_print) > 0:
            header_print(self.name)
            for string, args, kwargs in self.lines_to_print:
                string = self.writer_prefix + str(string)
                pretty_print(string, *args, **kwargs)
        self.lines_to_print = []
        
    return wrapper

# def agox_writer(func):
#     @functools.wraps(func)
#     def wrapper(self, state, *args, **kwargs):
#         if not self.use_counter:
#             header_print(self.name)
#         state = func(self, state, *args, **kwargs)
#         if self.use_counter and len(self.lines_to_print) > 0:
#             header_print(self.name)
#             for string, args, kwargs in self.lines_to_print:
#                 string = self.writer_prefix + str(string)
#                 pretty_print(string, *args, **kwargs)
#         self.lines_to_print = []
        
#     return wrapper

class Writer:

    def __init__(self, verbose=True, use_counter=True, prefix=''):
        self.verbose = verbose
        self.use_counter = use_counter
        self.lines_to_print = []
        self.writer_prefix = prefix

    def writer(self, string, *args, **kwargs):
        if self.verbose:
            if self.use_counter:
                self.lines_to_print.append((string, args, kwargs))
            else:
                string = self.writer_prefix + str(string)
                pretty_print(string, *args, **kwargs)

    def header_print(self, string):
        if self.verbose:
            header_print(string)

if __name__ == '__main__':

            
    version_string = get_git_revision_short_hash()

    print(version_string)




from ase import Atoms
import functools

def candidate_list_comprehension(func):
    @functools.wraps(func)
    def wrapper(self, possible_list, *args, **kwargs):
        if isinstance(possible_list, list):
            results = []
            for item in possible_list:
                if isinstance(item, Atoms):
                    results.append(func(self, item, *args, **kwargs))
                else:
                    raise TypeError('List must be list of Atoms (or Candidate) objects.')
            return results
        elif isinstance(possible_list, Atoms):
            return func(self, possible_list, *args, **kwargs)
        else:
            raise TypeError('Object must be of type Atoms (or Candidate).')
    return wrapper




import os
from abc import ABC, abstractmethod

import numpy as np
from ase.calculators.calculator import Calculator, all_changes

from agox.candidates.ABC_candidate import CandidateBaseClass
from agox.module import Module
from agox.observer import Observer
from agox.writer import Writer, agox_writer


class ModelBaseClass(Calculator, Observer, Writer, ABC):

    """Model Base Class implementation

    Parameters
    ----------
    database : AGOX Database obj
        If used for on-the-fly training on a database, this should be set
    verbose : int
        Verbosity of model logs.
    iteration_start_training : int
        When model is attached as observer it starts training after this number of
        iterations.
    update_period : int
        When model is attached as observer it updates every update_period
        iterations.
    record : set
        Training record.

    """

    def __init__(
        self,
        database=None,
        filter=None,
        order=0,
        verbose=True,
        use_counter=True,
        prefix="",
        iteration_start_training=0,
        update_period=1,
        surname="",
        gets={},
        sets={},
    ):
        """__init__ method for model base class

        If a database is supplied the model will attach itself as an observer on the database.
        If database is None the model needs to be trained manually.

        Parameters
        ----------
        database : AGOX Database obj

        order : int
            AGOX execution order
        verbose : int
            Verbosity
        use_counter : int
            Writer settings
        prefix : str
            Writer settings

        """
        Observer.__init__(self, order=order, surname=surname, gets=gets, sets=sets)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)
        Calculator.__init__(self)
        Module.__init__(self)

        self.filter = filter
        self._save_attributes = ["_ready_state"]

        self.verbose = verbose
        self.iteration_start_training = iteration_start_training
        self.update_period = update_period

        self.validation_data = []
        self._ready_state = False
        self._record = set()
        self.update = False

        self.add_observer_method(
            self.training_observer,
            gets=self.gets[0],
            sets=self.sets[0],
            order=self.order[0],
            handler_identifier="database",
        )

        if database is not None:
            self.attach_to_database(database)

    @property
    @abstractmethod
    def name(self):
        """str: Name of model. Must be implemented in child class."""
        pass

    @property
    @abstractmethod
    def implemented_properties(self):
        """:obj: `list` of :obj: `str`: Implemented properties.
        Available properties are: 'energy', 'forces', 'uncertainty'

        Must be implemented in child class.
        """
        pass

    @abstractmethod
    def predict_energy(self, atoms, **kwargs):
        """Method for energy prediction.

        Note
        ----------
        Always include **kwargs when implementing this function.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        float
            The energy prediction

        Must be implemented in child class.
        """
        pass

    @abstractmethod
    def train(self, training_data, **kwargs):
        """Method for model training.

        Note
        ----------
        Always include **kwargs when implementing this function.
        If your model is not trainable just write a method that does nothing

        Parameters
        ----------
        atoms : :obj: `list` of :obj: `ASE Atoms`
            List of ASE atoms objects or AGOX candidate objects to use as training data.
            All atoms must have a calculator with energy and other nesseary properties set, such that
            it can be accessed by .get_* methods on the atoms.


        Must be implemented in child class.

        """
        pass

    @property
    def ready_state(self):
        """bool: True if model has been trained otherwise False."""
        return self._ready_state

    @ready_state.setter
    def ready_state(self, state):
        self._ready_state = bool(state)

    def add_save_attributes(self, attribute):
        """Add attribute to save list.

        Parameters
        ----------
        attribute : str or list of str
            Name of attribute to add to save list.

        """
        if isinstance(attribute, str):
            self._save_attributes.append(attribute)
        else:
            self._save_attributes += attribute

    def remove_save_attributes(self, attribute):
        """Remove attribute from save list.

        Parameters
        ----------
        attribute : str or list of str
            Name of attribute to remove from save list.

        """
        if isinstance(attribute, str):
            self._save_attributes.remove(attribute)
        else:
            for a in attribute:
                self._save_attributes.remove(a)

    def reset_save_attributes(self):
        """Reset save list."""
        self._save_attributes = []

    @agox_writer
    @Observer.observer_method
    def training_observer(self, database, state):
        """Observer method for use with on-the-fly training based data in an AGOX database.

        Note
        ----------
        This implementation simply calls the train_model method with all data in the database

        Parameters
        ----------
        atoms : AGOX Database object
            The database to keep the model trained against

        Returns
        ----------
        None

        """
        iteration = self.get_iteration_counter()

        if iteration < self.iteration_start_training:
            return
        if (
            iteration % self.update_period != 0
            and iteration != self.iteration_start_training
        ):
            return

        data = database.get_all_candidates()
        self.train(data)

    def add_validation_data(self, data):
        """Add validation data to model.

        Parameters
        ----------
        data : :obj: `list` of :obj: `ASE Atoms`
            List of ASE atoms objects or AGOX candidate objects to use as validation data.
            All atoms must have a calculator with energy and other nesseary properties set, such that
            it can be accessed by .get_* methods on the atoms.

        """
        if isinstance(data, list):
            self.validation_data += data
        else:
            self.validation_data.append(data)

    def predict_forces(self, atoms, **kwargs):
        """Method for forces prediction.

        The default numerical central difference force calculation method is used, but
        this can be overwritten with an analytical calculation of the force.

        Note
        ----------
        Always include **kwargs when implementing this function.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        np.array
            The force prediction with shape (N,3), where N is len(atoms)

        """
        return self.predict_forces_central(atoms, **kwargs)

    def predict_forces_central(
        self, atoms, acquisition_function=None, d=0.001, **kwargs
    ):
        """Numerical cenral difference forces prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.
        acquisition_function : Acquisition function or None
            Function that takes evaluate acquisition function based on
            energy and uncertainty prediction. Used for relaxation in acquisition
            funtion if force uncertainties are not available.

        Returns
        ----------
        np.array
            The force prediction with shape (N,3), where N is len(atoms)

        """
        if acquisition_function is None:
            energy = lambda a: self.predict_energy(a)
        else:
            energy = lambda a: acquisition_function(
                *self.predict_energy_and_uncertainty(a)
            )

        e0 = energy(atoms) #self.predict_energy(atoms) 
        energies = []

        for a in range(len(atoms)):
            for i in range(3):
                new_pos = atoms.get_positions()  # Try forward energy
                new_pos[a, i] += d
                atoms.set_positions(new_pos)
                if atoms.positions[a, i] != new_pos[a, i]:  # Check for constraints
                    energies.append(e0)
                else:
                    energies.append(energy(atoms))
                    atoms.positions[a, i] -= d

                new_pos = atoms.get_positions()  # Try backwards energy
                new_pos[a, i] -= d
                atoms.set_positions(new_pos)
                if atoms.positions[a, i] != new_pos[a, i]:
                    energies.append(e0)
                else:
                    energies.append(energy(atoms))
                    atoms.positions[a, i] += d

        penergies = np.array(energies[0::2])  # forward energies
        menergies = np.array(energies[1::2])  # backward energies

        forces = ((menergies - penergies) / (2 * d)).reshape(len(atoms), 3)
        return forces

    def predict_uncertainty(self, atoms, **kwargs):
        """Method for energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        float
            The energy uncertainty prediction

        """
        warning.warn("Uncertainty is not implemented and will return 0.")
        return 0

    def predict_uncertainty_forces(self, atoms, **kwargs):
        """Method for energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        np.array
            The force uncertainty prediction with shape (N,3) with N=len(atoms)

        """
        warning.warn("Uncertainty is not implemented and will return 0.")
        return np.zeros((len(atoms), 3))

    def predict_energy_and_uncertainty(self, atoms, **kwargs):
        """Method for energy and energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        float, float
            The energy and energy uncertainty prediction

        """
        return self.predict_energy(atoms, **kwargs), self.predict_uncertainty(
            atoms, **kwargs
        )

    def predict_forces_and_uncertainty(self, atoms, **kwargs):
        """Method for energy and energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        np.array, np.array
            Forces and forces uncertainty. Both with shape (N, 3) with N=len(atoms).

        """
        return self.predict_forces(atoms, **kwargs), self.predict_forces_uncertainty(
            atoms, **kwargs
        )

    def converter(self, atoms, **kwargs):
        """Converts an ASE atoms object to a format that can be used by the model

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        object
            The converted object

        """
        return {}

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        """ASE Calculator calculate method

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for to predict properties of.
        properties : :obj: `list` of :obj: `str`
            List of properties to calculate for the atoms
        system_changes : ASE system_changes
            Which ASE system_changes to check for before calculation

        Returns
        ----------
        None
        """
        Calculator.calculate(self, atoms, properties, system_changes)

        E = self.predict_energy(self.atoms)
        self.results["energy"] = E

        if "forces" in properties:
            forces = self.predict_forces(self.atoms)
            self.results["forces"] = forces

    def validate(self, **kwargs):
        """Method for validating the model.

        Parameters
        ----------
        kwargs : dict
            Keyword arguments to pass to the validation method.

        Returns
        ----------
        dict
            Dictionary with validation results

        """
        if len(self.validation_data) == 0:
            return None

        E_true = []
        E_pred = []
        for d in self.validation_data:
            E_true.append(d.get_potential_energy())
            E_pred.append(self.predict_energy(d))

        E_true = np.array(E_true)
        E_pred = np.array(E_pred)

        return {
            "Energy MAE [eV]": np.mean(np.abs(E_true - E_pred)),
            "Energy RMSE [eV]": np.sqrt(np.mean((E_true - E_pred) ** 2)),
            "Max absolute energy error [eV]": np.max(np.abs(E_true - E_pred)),
            "Max relative energy error [%]": np.max((E_true - E_pred) / E_true)*100,
            "Min relative energy error [%]": np.min((E_true - E_pred) / E_true)*100,
        }

    def _training_record(self, data):
        """
        Record the training data.

        Parameters
        ----------
        data : list
            List of Atoms objects.

        """
        if not all([isinstance(d, CandidateBaseClass) for d in data]):
            return

        for d in data:
            self._record.add(d.cache_key)

        self.update = True

    def _get_new_data(self, data):
        """
        Get the new data.

        Parameters
        ----------
        data : list
            List of Atoms objects.

        Returns
        -------
        list
            List of new Atoms objects.

        list
            List of old Atoms objects.

        """
        if not all([isinstance(d, CandidateBaseClass) for d in data]):
            return data, []

        new_data = []
        old_data = []
        for d in data:
            if d.cache_key in self._record:
                old_data.append(d)
            else:
                new_data.append(d)
        return new_data, old_data

    @agox_writer
    def print_model_info(self, validation=None, **kwargs):
        """Prints model information

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments to pass to the model

        Returns
        ----------
        None
        """
        model_info = self.model_info(**kwargs)
        if validation is not None:
            model_info.append("------ Validation Info ------")
            model_info.append(
                "Validation data size: {}".format(len(self.validation_data))
            )
            for key, val in validation.items():
                model_info.append("{}: {:.3}".format(key, val))

        for s in model_info:
            self.writer(s)

    def model_info(self, **kwargs):
        """Returns model information

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments to pass to the model

        Returns
        ----------
        list of str
            The model information
        """
        return ["No model information available."]

    @agox_writer
    def save(self, path="model.h5"):
        """
        Save the model as a pickled object.

        Parameters
        ----------
        path : str, optional
            Path to save the model to. The default is 'model.h5'.
        """
        import h5py

        with h5py.File(path, "w") as f:
            for key in self._save_attributes:
                obj = self
                for k in key.split("."):
                    data = getattr(obj, k)
                    obj = data
                if data is not None:
                    f.create_dataset(key, data=data)

        self.writer("Saving model to {}".format(path))

    @agox_writer
    def load(self, path):
        """
        Load a pickle

        Parameters
        ----------
        path : str
            Path to a saved model.

        Returns
        -------
        model-object
            The loaded model object.
        """
        assert os.path.exists(path), "Path does not exist"
        import h5py

        self.writer("Loading model from {}".format(path))
        with h5py.File(path, "r") as f:
            for key in self._save_attributes:
                obj = self
                for k in key.split(".")[:-1]:
                    obj = getattr(obj, k)

                k = key.split(".")[-1]
                
                if key in f:
                    value = f[key][()]
                    setattr(obj, k, value)

        self.writer("Model loaded")

    def attach_to_database(self, database):
        from agox.databases.ABC_database import DatabaseBaseClass

        assert isinstance(database, DatabaseBaseClass)
        print(f"{self.name}: Attaching to database: {database}")
        self.attach(database)



from typing import Optional, Tuple, Union

import numpy as np
from ase.atoms import Atoms
from numpy.typing import NDArray

from agox.candidates.standard import StandardCandidate
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass


class SurfaceCenteringPostprocess(PostprocessBaseClass):

    """Centers a candidate object to the middle of the cell, taking into
    account adsorption on top of a periodic surface on the xy plane.

    Parameters
    ----------
    template_size : Union[Tuple[int, int], Tuple[int, int, int], NDArray]
        The size of the template given in units of the minimal unit cell. This
        is equivalent to the `size` parameter provided to surface building
        methods from `ase.build`.
    desired_location : Optional[Union[Tuple[int, int], NDArray]], optional
        The desired location of the searched structure in units of the minimal
        unit cell, by default None. If None, take half of `template_size`
        (rounded down).
    """

    name = 'SurfaceCenteringPostprocess'

    def __init__(self,
                 template_size: Union[Tuple[int, int], Tuple[int, int, int], NDArray],
                 desired_location: Optional[Union[Tuple[int, int], NDArray]] = None,
                 **kwargs):
        super().__init__(**kwargs)

        template_size = np.array(template_size, dtype=int)

        if template_size.shape not in [(2,), (3,)]:
            raise ValueError(f'Invalid shape for template_size; is {template_size.shape}, must be (2,) or (3,).')

        self.template_size = template_size[:2]  # we only care about xy

        if desired_location is not None:
            self.desired_location = np.array(desired_location, dtype=int)
        else:
            self.desired_location = self.template_size // 2

        if self.desired_location.shape != (2,):
            raise ValueError(f'Invalid shape for desired_location; if {self.desired_location.shape}, must be (2,).')

        if np.any((self.desired_location < 0) | (self.desired_location >= self.template_size)):
            raise ValueError(f'Invalid desired_location: {list(self.desired_location)}. '
                             f'Must be between [0, 0] and {list(self.template_size - 1)}.')

    @PostprocessBaseClass.immunity_decorator
    def postprocess(self, candidate: StandardCandidate) -> StandardCandidate:
        """Move the searched structure on the surface if necessary.

        Parameters
        ----------
        candidate : StandardCandidate
            The candidate object containing a surface-supported set of atoms.

        Returns
        -------
        StandardCandidate
            A copy of the candidate, with the searched structure translated to
            the desired position.
        """
        original_candidate = candidate
        candidate = candidate.copy()

        atoms = Atoms(candidate)

        cell_xy = candidate.get_cell()[:2, :2]
        uc_xy = cell_xy / self.template_size[:, np.newaxis]

        optimize_indices = candidate.get_optimize_indices()

        com_xy = atoms[optimize_indices].get_center_of_mass()[:2]
        com_uc = np.linalg.solve(uc_xy.T, com_xy)

        offset_uc = np.floor(com_uc + 0.5) - self.desired_location
        offset_xy = np.dot(offset_uc, uc_xy)

        candidate.positions[optimize_indices, :2] -= offset_xy

        original_candidate.copy_calculator_to(candidate)

        return candidate



import pytest
import numpy as np
from agox.candidates import CandidateBaseClass
from agox.helpers.confinement import Confinement
from agox.generators import RandomGenerator, RattleGenerator, ReplaceGenerator, CenterOfGeometryGenerator, SamplingGenerator

generator_params = [
    (RandomGenerator, (), {}),
    (RandomGenerator, (), {'may_nucleate_at_several_places': True, 'c2':2.5}),
    (RattleGenerator, (), {}),
    (RattleGenerator, (), {'n_rattle': 1}),
    (RattleGenerator, (), {'rattle_amplitude': 1}),
    (ReplaceGenerator, (), {}),
    (ReplaceGenerator, (), {'n_replace': 3}),
    (ReplaceGenerator, (), {'amplitude': 1}),
    (CenterOfGeometryGenerator, (), {}),
    (SamplingGenerator, (), {}),

]
@pytest.fixture(scope='module', params=generator_params)
def candidate(request, sampler, test_environment):


    generator = request.param[0](*request.param[1],
                                 **test_environment.get_confinement(), 
                                 **request.param[2])
    
    candidate = None
    attempt = 0
    while candidate is None and attempt < 10:
        candidate = generator(sampler, test_environment)[0]
        attempt += 1
    return candidate

def test_output_type(candidate):
    assert isinstance(candidate, CandidateBaseClass)

def test_atomic_formula(candidate, test_environment):
    candidate_numbers = np.sort(np.array(candidate.get_atomic_numbers()))
    env_numbers = np.sort(np.array(test_environment.get_all_numbers()))

    assert (candidate_numbers == env_numbers).all()

def test_confinement(candidate, test_environment):
    indices = test_environment.get_missing_indices()
    positions = candidate.positions[indices, :]
    confinement = Confinement(**test_environment.get_confinement())
    assert confinement.check_confinement(positions).all()

def test_template(candidate, test_environment):
    template = test_environment.get_template()
    assert (candidate.positions[0:len(template)] == template.positions).all()

def test_cell(candidate, test_environment):
    assert (candidate.cell == test_environment.get_template().get_cell()).all()


from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
import numpy as np


class RattleGenerator(GeneratorBaseClass):

    """
    Rattles atoms in a seed structure, e.g. perturbs atoms in random directions.

    Parameters
    -----------
    n_rattle : int
        Medium number of atoms to rattle.
    rattle_amplitude : float
        Maximum distance to rattle atoms.
    replace : bool
        ?
    """

    name = "RattleGenerator"

    def __init__(self, n_rattle=3, rattle_amplitude=3, replace=True, **kwargs):
        super().__init__(replace=replace, **kwargs)
        self.n_rattle = n_rattle
        self.rattle_amplitude = rattle_amplitude

    def _get_candidates(self, candidate, parents, environment):
        indices_to_rattle = self.get_indices_to_rattle(candidate)

        for i in indices_to_rattle:
            for _ in range(100):  # For atom i attempt to rattle up to 100 times.
                radius = self.rattle_amplitude * np.random.rand() ** (1 / self.dimensionality)
                displacement = self.get_displacement_vector(radius)
                suggested_position = candidate.positions[i] + displacement

                # Check confinement limits:
                if not self.check_confinement(suggested_position).all():
                    continue

                # Check that suggested_position is not too close/far to/from other atoms
                # Skips the atom it self.
                if self.check_new_position(
                    candidate,
                    suggested_position,
                    candidate[i].number,
                    skipped_indices=[i],
                ):
                    candidate[i].position = suggested_position
                    break

        return [candidate]

    def get_indices_to_rattle(self, candidate):
        # Establish indices_to_rattle as the indices of the atoms that should be rattled
        template = candidate.get_template()
        n_template = len(template)
        n_total = len(candidate)
        n_non_template = n_total - n_template
        probability = self.n_rattle / n_non_template
        indices_to_rattle = np.arange(n_template, n_total)[np.random.rand(n_non_template) < probability]
        indices_to_rattle = np.random.permutation(indices_to_rattle)
        if len(indices_to_rattle) == 0:
            indices_to_rattle = [np.random.randint(n_template, n_total)]
        # now indices_to_rattle is ready to use
        return indices_to_rattle

    def get_number_of_parents(self, sampler):
        return 1



from typing import List

import numpy as np
from ase import Atoms

from agox.utils.filters.ABC_filter import FilterBaseClass


class EnergyFilter(FilterBaseClass):
    """
    Filter that removes all structures with energy above a certain threshold.

    """

    name = "EnergyFilter"

    def __init__(self, delta_E: float = 1000, **kwargs):
        """
        Parameters
        ----------
        max_energy: float
            Maximum energy above minimum energy.
        """
        super().__init__(**kwargs)
        self.delta_E = delta_E

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        Es = np.array([a.get_potential_energy() for a in atoms])
        E_min = np.min(Es)
        E_boundary = E_min + self.delta_E
        # get indicies where Es is greater than E_boundary
        indexes = np.array([i for i, E in enumerate(Es) if E < E_boundary])

        return indexes



import numpy as np
from agox.generators.complementary_energy.attractor_methods.ABC_attractors import AttractorMethodBaseClass
from sklearn.cluster import KMeans

class AttractorKmeans(AttractorMethodBaseClass):
    """
    Method for picking attractors from another structures using K-means.

    Parameters
    -----------
    descriptor: Local descriptor
        Used to calculate local features
    number_of_structures: int
        Number of structures from database to use when creating attractors
    possible_number_of_attractors: int or list with two ints
        If int, then that exact amount of attractors will be created. If a list,
        the number of attractors will be a random number between the two given numbers
    attractors_from_template: Bool
        Whether attractors can be picked as feature vectors of template atoms
    predefined_attractors: List of attractors
        If not None, the specified attractors will be given to the CE generator
    order: int
        When to update the structures by checking the database
    """

    name = 'AttractorKmeans'

    def __init__(self, descriptor, number_of_structures = 5, possible_number_of_attractors = [3, 6], attractors_from_template = True, predefined_attractors = None, order = 5):
        super().__init__(order = order)
        self.descriptor = descriptor
        self.number_of_structures = number_of_structures
        self.possible_number_of_attractors = possible_number_of_attractors
        self.attractors_from_template = attractors_from_template
        self.predefined_attractors = predefined_attractors

    def get_attractors(self, structure):

        if self.predefined_attractors is not None:
            return self.predefined_attractors
        
        template = structure.get_template()
        n_template = len(template)

        if isinstance(self.possible_number_of_attractors, int):
            number_of_attractors = self.possible_number_of_attractors
        else:
            number_of_attractors = np.random.randint(self.possible_number_of_attractors[0], self.possible_number_of_attractors[1])

        if len(self.structures) <= self.number_of_structures:
            database_structure_index = range(len(self.structures))
        else:
            database_structure_index = np.random.choice(len(self.structures), size = self.number_of_structures, replace = False)
    
        structures_from_database = [self.structures[i] for i in database_structure_index]

        list_of_features = []
        for struc in structures_from_database:
            features = self.descriptor.get_features(struc)
            if not self.attractors_from_template:
                features = features[n_template:]
            list_of_features.append(features)

        list_of_features = np.array(list_of_features).reshape(len(structures_from_database) * len(list_of_features[0]), -1)

        kmeans = KMeans(n_clusters = number_of_attractors).fit(list_of_features)
        attractors = kmeans.cluster_centers_

        return attractors


from ase.calculators.calculator import Calculator, all_changes, all_properties
import os
from pathlib import Path

class SubprocessGPAW(Calculator):

        """
        GPAW calculator that uses a subprocess to run the calculation.

        Parameters
        -------------
        ncores (int): The number of cores to use. 
            If None, the number of cores is determined by the number of cores available.

        **kwargs:
            Additional keyword arguments to pass to the GPAW for calculation settings. 
            Eg. 
                kwargs = {
                    'mode': {'name': 'pw', 'ecut': cutoff},
                    'xc': 'PBE',
            Is a planewave PBE calculation. 
            The same dict can be passed directly to the normal GPAW calculator, to exactly 
            reproduce the calculation settings.
        """

        implemented_properties = ['energy', 'forces', 'stress']

        def __init__(self, ncores=None, log_directory='gpaw_logs/', **kwargs):
            super().__init__()
            self.kwargs = kwargs

            if ncores is None:
                ncores = len(os.sched_getaffinity(0))
            self.ncores = ncores
            self.count = 0

            self.log_directory = Path(log_directory)
            if not self.log_directory.exists():
                self.log_directory.mkdir(parents=True)

        def calculate(self, atoms=None, properties=all_properties, system_changes=all_changes):
            from ase.calculators.subprocesscalculator import gpaw_process # Because of ASE versions.

            super().calculate(atoms, properties, system_changes)

            # Update the log file name.
            kwargs = self.kwargs.copy()
            if kwargs.get('txt', '-') != '-':
                name = Path(kwargs['txt']).stem + f"_{self.count}.txt"
                kwargs['txt'] = str(self.log_directory / name)
            
            # Run the calculation.
            with gpaw_process(ncores=self.ncores, **kwargs) as gpaw:
                gpaw._run_calculation(self.atoms.copy(), properties, system_changes)
                results = gpaw.protocol.recv()

            self.results.update(results)
            self.count += 1


import pytest
import numpy as np

from ase.calculators.lj import LennardJones
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.models.descriptors import SOAP, Fingerprint
from agox.utils.sparsifiers.CUR import CUR
from agox.models.GPR.priors import Repulsive
from ase import Atoms

def model_factory(sparsification_schedule=None):
    # Setup a model: 
    descriptor = SOAP.from_species(['H'], periodic=False, r_cut=7)
    kernel = C(1, (1, 1))*RBF(10, (20, 20))
    sparsifier = CUR(5)

    general_kwargs = dict(
        descriptor=descriptor,
        kernel=kernel,
        sparsifier=sparsifier,
        noise=1,
        prior=Repulsive(ratio=0.7),
        use_ray=False,
        n_optimize=0, 
        sparsification_schedule=sparsification_schedule)

    model = SparseGPR(**general_kwargs)
    return model

def pair_potential_data():
    d = np.linspace(0.88, 1.7, 20)
    data = []
    for d_ in d:
        atoms = Atoms('HH', positions=[[0, 0, 0], [0, 0, d_]])
        atoms.calc = LennardJones()
        atoms.get_potential_energy()
        data.append(atoms)

    return data


@pytest.mark.ray
def test_sparse_model_schedule_1():
    # Setup a model:
    model = model_factory()
    data = pair_potential_data()

    # Train the model and it selects sparse points: 
    model.train(data)

    # Check that the model has selected the correct number of sparse points:
    assert len(model.Xm) == 5

    # Reset the model and check that it does sparsification:
    model.Xm = model.Xm[0:1, :]
    model.train(data)
    assert len(model.Xm) == 5

@pytest.mark.ray
def test_sparse_model_schedule_2():

    def sparsification_schedule(*args, **kwargs):
        return False

    # Setup a model:
    model = model_factory(sparsification_schedule=sparsification_schedule)

    data = pair_potential_data()

    # Train the model and it selects sparse points:
    model.train(data)

    # Check that the model has selected the correct number of sparse points:
    assert len(model.Xm) == 5

    # Reset the model and check that it doesn't do sparsification:
    model.iteration_counter = 1
    model.Xm = model.Xm[0:1, :]
    model.train(data)
    assert len(model.Xm) == 1



import numpy as np 

def atomic_numerical_derivative(func, atoms, delta=1e-3):

    N_atoms = len(atoms)

    f_output_shape = func(atoms).shape
    F = np.zeros((N_atoms, 3, *f_output_shape))

    stencil = [-2, -1, 0, 1, 2]

    for a in range(N_atoms):
        for d in range(3):        
            f = {}
            for s in stencil: 
                atoms_tmp = atoms.copy()
                atoms_tmp.positions[a, d] += delta * s
                f[s] = func(atoms_tmp)

            F[a, d] = (1*f[-2]-8*f[-1]+0*f[0]+8*f[1]-1*f[2]) / (12 * delta)

    return -F






import numpy as np
import pytest
from ase.io import read
from agox.utils.sparsifiers import RandomSparsifier, CUR, MBkmeans


@pytest.mark.parametrize('sparsifier_dict', [{'sparsifier':RandomSparsifier, 'kwargs':{}}, 
                                             {'sparsifier':CUR, 'kwargs':{}},
                                             {'sparsifier':MBkmeans, 'kwargs':{'exact_points':True}}
])
def test_sparsifier(sparsifier_dict):
    # Random feautre matix:
    X = np.random.rand(1000, 20)

    # Setup a sparsifier
    sparsifier = sparsifier_dict['sparsifier'](m_points=10, **sparsifier_dict['kwargs'])

    # Sparsify the structures
    Xm, indices = sparsifier.sparsify(X)

    print('Before sparsification: ', X.shape)
    print('After sparsification: ', Xm.shape)

    # Assert that the shapes make sense.
    assert Xm.shape[0] == 10
    assert Xm.shape[1] == 20
    assert len(indices) == 10

    # Assert that using the indices to select rows from X gives the same result as Xm
    assert np.allclose(Xm, X[indices, :])











from agox.collectors.ABC_collector import CollectorBaseClass
from timeit import default_timer as dt

class StandardCollector(CollectorBaseClass):

    """
    Makes a number of candidates from each generator serially.

    The StandardCollector allows for either constant or iteration dependent 
    control of the number of candidates produced by each given generator. 

    Parameters
    ----------
    num_candidates : list or dict
        If a list it have the same length as the generators list containing 
        integers dictating how many candidates are produced by each 
        generator in the same order as generators are given. E.g. 
            num_candidates = [10, 5, 5]
        If a dict it must have integer keys that and 0 must be a key. The 
        values of the dict are lists like a list is passed, the keys dictate 
        at which episode the corresponding list is used, e.g
            num_candidates = {0:[10, 5, 5], 10:[10, 10, 10]}
    """

    name = 'StandardCollector'

    def __init__(self, num_candidates, **kwargs):
        super().__init__(**kwargs)
        self.num_candidates = num_candidates

    def get_number_of_candidates(self):
        if type(self.num_candidates) == list:
            return self.num_candidates
        elif type(self.num_candidates) == dict:
            return self.get_number_of_candidates_for_iteration()
            
    def get_number_of_candidates_for_iteration(self):
        # self.num_candidates must have this form: {0: [], 500: []}
        keys = list(self.num_candidates.keys())
        keys.sort()
        iteration = self.get_iteration_counter()
        if iteration is None:
            iteration = 0

        num_candidates = self.num_candidates[0] # yes, it must contain 0
        # now step through the sorted list (as long as the iteration is past the key) and extract the num_candidates
        # the last one extracted will be the most recent num_candidates enforced and should apply to this iteration
        for k in keys:
            if iteration < k:
                break
            num_candidates = self.num_candidates[k]
        return num_candidates

    def make_candidates(self):

        all_candidates = []
        # If sampler is empty and initialized, then use fallback generator
        if len(self.sampler) == 0 and self.sampler.initialized:
            fallback_generator = self.get_fallback_generator()
            num_candidates = sum(self.get_number_of_candidates())
            for sample in range(num_candidates):
                candidates = fallback_generator(sampler=self.sampler, environment=self.environment)

                for candidate in candidates:
                    all_candidates.append(candidate)

        else:
            for generator, num_candidates in zip(self.generators, self.get_number_of_candidates()):
                for sample in range(num_candidates):
                    candidates = generator(sampler=self.sampler, environment=self.environment)

                    for candidate in candidates:
                        all_candidates.append(candidate)
                        
        return all_candidates
            




from agox.observer import Observer
from agox.writer import Writer, agox_writer
from abc import ABC, abstractmethod

class AttractorMethodBaseClass(ABC, Observer, Writer):
    """
    Baseclass for attractors methods

    Parameters
    -----------
    order: int
        When to update the structures by checking the database
    """

    dynamic_attributes = ['structures']

    def __init__(self, order = 10):
        Observer.__init__(self, order = order)
        Writer.__init__(self)
        self.add_observer_method(self.update, order=self.order[0], 
            sets={}, gets={})
        
        self.structures = []
    
    @agox_writer
    @Observer.observer_method
    def update(self, database, state):
        self.structures = database.get_all_candidates()

    @abstractmethod
    def get_attractors(self, structure):
        pass


from copy import deepcopy

import numpy as np
from ase import Atoms
from ase.constraints import FixInternals
from ase.data import covalent_radii
from scipy.spatial.distance import cdist

from agox.generators.block_generators.ABC_block import BlockGeneratorBaseClass


class RattleBlockGenerator(BlockGeneratorBaseClass):
    name = "RattleBlockGenerator"

    def __init__(
        self,
        blocks_to_rattle=2,
        rotation_min=0,
        rotation_max=15,
        displacement_min=0,
        displacement_max=2,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.blocks_to_rattle = blocks_to_rattle
        self.rotation_min = rotation_min
        self.rotation_max = rotation_max
        self.displacement_min = displacement_min
        self.displacement_max = displacement_max

    def _get_candidates(self, candidate, parents, environment):
        # Block information:
        inter_block_indices = parents[0].meta_information.get("block_indices", None)
        if inter_block_indices is None:
            return []
        N_blocks = len(inter_block_indices)
        block_probs = np.ones(N_blocks) / N_blocks

        # Number of blocks to rattle:
        blocks_to_rattle = self.blocks_to_rattle  # Simple for now.
        atleast_one = False

        for n in range(blocks_to_rattle):
            progress = False
            attempt = 0
            while not progress and attempt <= 1000:
                attempt += 1

                # Pick a block to move:
                block_index = np.random.choice(
                    np.arange(N_blocks), size=1, p=block_probs
                )[0]
                inter_indices = np.array(inter_block_indices[block_index], dtype=int)
                block = self.extract_block(candidate, inter_indices)

                self.rattle_block(block)  # Now we have a rattled block. In place.

                # CHeck if the block obeys confinement:
                if not self.check_confinement(block.positions).any():
                    continue

                if self.check_distances(candidate, block, inter_indices, present=True):
                    self.block_move(candidate, block, inter_indices)

                    # Renormalize distribution:
                    block_probs[block_index] = 0
                    if np.sum(block_probs) > 0:
                        block_probs /= np.sum(block_probs)

                    atleast_one = True
                    progress = True

        if not atleast_one:
            return []

        blocks_used = parents[0].meta_information.get("blocks_used")
        candidate.add_meta_information("blocks_used", blocks_used)
        candidate.add_meta_information(
            "block_indices", inter_block_indices
        )

            
        return [candidate]

    def rattle_block(self, block):
        for a in range(3):
            axis = np.zeros(3)
            axis[a] = 1
            angle = (
                np.random.rand() * (self.rotation_max - self.rotation_min)
                + self.rotation_min
            )
            sign = np.random.choice([-1, 1], size=1)[0]
            block.rotate(sign * angle, axis, center="COP")

        radius = (
            np.random.rand() * (self.displacement_max - self.displacement_min)
            + self.displacement_min
        )
        d = self.get_displacement_vector(radius)
        block.positions += d

    def get_number_of_parents(self, sampler):
        return 1



import numpy as np
from agox.acquisitors.ABC_acquisitor import AcquisitorBaseClass

class MetaInformationAcquisitor(AcquisitorBaseClass):

    name = 'MetaInformationAcquisitor'

    def __init__(self, meta_key, mode='min', **kwargs):
        super().__init__(**kwargs)
        self.meta_key = meta_key
        self.mode = mode

        assert mode in ['max', 'min'] # Either maximize or minimize according to the meta key.

    def calculate_acquisition_function(self, candidates):
        sign = 1 if self.mode == 'min' else -1
        fitness = np.zeros(len(candidates))
        
        for i, candidate in enumerate(candidates):
            fitness[i] = sign * candidate.get_meta_information(self.meta_key)

        return fitness
    
    







import numpy as np
from pathlib import Path
from agox.test.test_utils import compare_candidates
from agox.test.test_utils import TemporaryFolder, test_folder_path, check_folder_is_empty
from importlib import import_module
from agox.databases import Database
import os

def agox_test_run(mode, tmp_path, cmd_options, full_compare=True):

    # Find the path to the current test file
    repo_path = Path(test_folder_path).parent.parent
    env_vars = os.getenv('PYTEST_CURRENT_TEST').split('::')[0]
    test_path = Path(env_vars).parent
    expected_path = repo_path / test_path / 'expected_outputs' / f'{mode}_test'

    # Find the script path:
    rel_script_path = test_path / f'script_{mode}.py'
    full_script_path = repo_path / rel_script_path
    script_import = str(rel_script_path).replace('/', '.').replace('.py', '')


    # Check mode & tolerance
    create_mode = cmd_options['create_mode']
    test_mode = not create_mode
    tolerance = cmd_options['tolerance']

    if create_mode:
        tmp_path = expected_path
        check_folder_is_empty(tmp_path)

    # Prints for helping later:
    print('====== Test information ======')
    print(f'\tExpected results path: \n\t\t{expected_path}')
    print(f'\tScript path: \n\t\t{full_script_path}')
    print(f'\tScript import: \n\t\t{script_import}')
    print('')

    with TemporaryFolder(tmp_path):
        # This loads the database file from the script file. 
        # This means that the documentation can link to this run-file.
        database = import_module(script_import).database

        if test_mode: 
            expected_database = Database(f'{expected_path}/db0.db')
            compare_runs(database, expected_database, tolerance, full_compare=full_compare)

def compare_runs(database, expected_database, tolerance, full_compare=True):
    test_candidates = database.get_all_candidates()
    test_energies = database.get_all_energies()

    # Saved database:
    expected_database.restore_to_memory()
    expected_candidates = expected_database.get_all_candidates()
    expected_energies = expected_database.get_all_energies()

    np.testing.assert_allclose(expected_energies, test_energies, **tolerance)
    
    if full_compare:
        for candidate, expected_candidate in zip(test_candidates, expected_candidates):
            assert compare_candidates(candidate, expected_candidate, tolerance), f'Candidates dont match. {candidate.positions = } {expected_candidate.positions = }'

        assert len(expected_candidates) == len(test_candidates), 'Different numbers of candidates.'

        expected_positions = np.array([atoms.get_positions() for atoms in expected_candidates])
        test_positions = np.array([atoms.get_positions() for atoms in test_candidates])

        np.testing.assert_allclose(expected_energies, test_energies, **tolerance)
        np.testing.assert_allclose(expected_positions, test_positions, **tolerance)





from typing import Tuple

import numpy as np
from sklearn.gaussian_process.kernels import RBF as sklearn_RBF
from sklearn.gaussian_process.kernels import \
    ConstantKernel as sklearn_ConstantKernel
from sklearn.gaussian_process.kernels import Product as sklearn_Product
from sklearn.gaussian_process.kernels import Sum as sklearn_Sum
from sklearn.gaussian_process.kernels import WhiteKernel as sklearn_WhiteKernel
from sklearn.gaussian_process.kernels import DotProduct as sklearn_DotProduct



class Kernel:

    def __add__(self, b):
        if not isinstance(b, Kernel):
            return Sum(self, Constant(b))
        return Sum(self, b)

    def __radd__(self, b):
        if not isinstance(b, Kernel):
            return Sum(Constant(b), self)
        return Sum(b, self)

    def __mul__(self, b):
        if not isinstance(b, Kernel):
            return Product(self, Constant(b))
        return Product(self, b)

    def __rmul__(self, b):
        if not isinstance(b, Kernel):
            return Product(Constant(b), self)
        return Product(b, self)

    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:

        dk_dy = np.zeros((X.shape[0], Y.shape[1]))

        Y0 = Y.copy()

        for i in range(Y.shape[1]):
            Yp = Y0.copy()
            Yp[:, i] += 1e-5
            Ym = Y0.copy()
            Ym[:, i] -= 1e-5
            dk_dy[:, i] = ((self(X, Yp) - self(X, Ym)) / (2 * 1e-5)).squeeze()

        return dk_dy


class RBF(Kernel, sklearn_RBF):
    def __init__(
        self,
        length_scale: float = 1.0,
        length_scale_bounds: Tuple[float, float] = (1e-1, 1e3),
    ) -> None:
        sklearn_RBF.__init__(
            self, length_scale=length_scale, length_scale_bounds=length_scale_bounds
        )

    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        if np.ndim(Y) == 1:
            Y = Y[np.newaxis, :]

        K = self(X, Y)
        dK_dd = -1 / (2 * self.length_scale**2) * K
        dd_df = -2 * (X - Y)
        dk_df = np.multiply(dK_dd, dd_df)

        return dk_df


class Constant(Kernel, sklearn_ConstantKernel):
    def __init__(
        self,
        constant_value: float = 1.0,
        constant_value_bounds: Tuple[float, float] = (1e-1, 1e5),
    ) -> None:
        sklearn_ConstantKernel.__init__(
            self,
            constant_value=constant_value,
            constant_value_bounds=constant_value_bounds,
        )

    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        return 0


class Noise(Kernel, sklearn_WhiteKernel):
    def __init__(
        self,
        noise_level: float = 0.01,
        noise_level_bounds: Tuple[float, float] = (0.01, 0.01),
    ) -> None:
        sklearn_WhiteKernel.__init__(
            self, noise_level=noise_level, noise_level_bounds=noise_level_bounds
        )

    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        return 0


class Sum(Kernel, sklearn_Sum):
    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        return self.k1.get_feature_gradient(X, Y) + self.k2.get_feature_gradient(X, Y)


class Product(Kernel, sklearn_Product):
    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        K1 = self.k1(X, Y)
        K2 = self.k2(X, Y)
        grad_K1 = self.k1.get_feature_gradient(X, Y)
        grad_K2 = self.k2.get_feature_gradient(X, Y)

        return K1 * grad_K2 + grad_K1 * K2


class DotProduct(Kernel, sklearn_DotProduct):

    # def __init__(self, normalize=False, **kwargs):
    #     super().__init__(**kwargs)
    #     self.normalize = normalize

    def __call__(self, X, Y=None, eval_gradient=False):
        return super().__call__(X, Y)
    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        return super().get_feature_gradient(X, Y)


from sklearn.gaussian_process.kernels import RationalQuadratic as sklearn_RationalQuadratic

class RationalQuadratic(Kernel, sklearn_RationalQuadratic):

    def get_feature_gradient(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        return super().get_feature_gradient(X, Y)

        # from scipy.spatial.distance import cdist

        # if np.ndim(Y) == 1:
        #     Y = Y[np.newaxis, :]

        # distance_squared = cdist(X, Y, metric='sqeuclidean')

        # A = 1 + distance_squared / (2 * self.alpha * self.length_scale ** 2)

        # derivative = 2 * self.alpha * A ** (-self.alpha - 1) * (Y - X)

        # return derivative



from agox.generators.ABC_generator import GeneratorBaseClass

class SamplingGenerator(GeneratorBaseClass):

    """
    Samples without pertubation from a sampler.

    The generator just returns a member of the sampler, which can then be processed by any
    postprocessors, such as relaxation in an updated model.
    """

    name = 'SamplingGenerator'

    def __init__(self, replace=True, **kwargs):
        super().__init__(replace=replace,**kwargs)

    def _get_candidates(self, candidate, parents, environment):

        candidates = parents

        if len(candidates) == 0:
            return []
        
        return candidates

    def get_number_of_parents(self, sampler):
        return len(sampler)



import numpy as np
import pytest 
from agox.environments import Environment
from agox.samplers import MetropolisSampler
from ase import Atoms
from agox.generators import RandomGenerator, RattleGenerator, CenterOfGeometryGenerator, ReplaceGenerator, ReuseGenerator, SamplingGenerator, PermutationGenerator
from agox.candidates import CandidateBaseClass

from agox.test.test_utils import environment_and_dataset, TemporaryFolder

@pytest.mark.parametrize('generator_class', [RandomGenerator, RattleGenerator, ReplaceGenerator, CenterOfGeometryGenerator, SamplingGenerator])
class TestGenerator:

    def assertions(self, candidates, environment, sampler):
        for candidate in candidates:
            assert issubclass(candidate.__class__, CandidateBaseClass)
            assert len(candidate) == len(environment.get_all_numbers())
            assert (candidate.cell == environment.get_template().get_cell()).all()

    def setup_generator(self, generator_class, environment, **kwargs):
        return generator_class(environment=environment, **environment.get_confinement(), **kwargs)

    def setup_sampler(self, dataset):
        sampler = MetropolisSampler()
        sampler.sample = [dataset[0]]
        return sampler

    def test_generators(self, generator_class, environment_and_dataset, tmp_path):
        with TemporaryFolder(tmp_path):
            environment, dataset = environment_and_dataset
            
            generator = self.setup_generator(generator_class, environment)
            sampler = self.setup_sampler(dataset)
            candidates = [None]
            for i in range(1):
                candidates = generator(sampler, environment)
                if not candidates[0] == None:
                    break
            self.assertions(candidates, environment, sampler)

            assert (tmp_path / f'confinement_plot_{generator.name}.png').is_file()



from agox.test.run_tests.run_utils import agox_test_run
import pytest

@pytest.mark.ray
@pytest.mark.parametrize("mode", ["gofee_cluster_emt", "gofee_surface_emt", "gofee_bulk_emt"])
def test_gofee(mode, tmp_path, cmd_options, ray_fix):
    agox_test_run(mode, tmp_path, cmd_options)



from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
from ase import Atoms
import numpy as np

from scipy.spatial.distance import cdist

class RandomGenerator(GeneratorBaseClass):

    """
    Creates structures without a seed and with minimal bias.

    Parameters
    -----------
    contiguous : bool
        If False, the generator may place atoms at several places on a template. If True, 
        any placed atom must be placed close to an already placed atom.
        Replaces 'may_nucleate_at_several_places' which is deprecated.
    replace : bool
        ?
    """

    name = 'RandomGenerator'

    def __init__(self, contiguous=True, may_nucleate_at_several_places=None, replace=True, **kwargs):
        super().__init__(replace=replace, **kwargs)
        self.contiguous = contiguous

        if may_nucleate_at_several_places is not None:
            DeprecationWarning(f"""'may_nucleate_at_several_places' is deprecated and will be removed. Use 'contiguous' instead.
                               Setting 'contiguous' to {not may_nucleate_at_several_places} which is the inverse of the 
                               supplied value for 'may_nucleate_at_several_places'.""")
            self.contiguous = not may_nucleate_at_several_places

    def _get_candidates(self, candidate, parents, environment):

        template = environment.get_template()
        numbers_list = environment.get_numbers()
        len_of_template = len(template)

        while len(numbers_list) > 0:
            np.random.shuffle(numbers_list)
            atomic_number = numbers_list[0]
            numbers_list = numbers_list[1:]

            placing_first_atom = (len(candidate) == len_of_template)

            for _ in range(100):
                if placing_first_atom or not self.contiguous: # If the template is completely empty. 
                    suggested_position = self.get_box_vector()
                else: # Pick only among atoms placed by the generator. 
                    placed_atom = candidate[np.random.randint(len_of_template,len(candidate))]
                    suggested_position = placed_atom.position.copy()
                    # Get a vector at an appropriate radius from the picked atom. 
                    vec = self.get_sphere_vector(atomic_number, placed_atom.number) 
                    suggested_position += vec

                if not self.check_confinement(suggested_position).all():
                    build_succesful = False
                    continue
                
                # Check that suggested_position is not too close/far to/from other atoms
                if self.check_new_position(candidate, suggested_position, atomic_number) or len(candidate) == 0:
                    build_succesful = True
                    candidate.extend(Atoms(numbers=[atomic_number], positions=[suggested_position]))
                    break
                else:
                    build_succesful = False

            if not build_succesful:
                self.writer('RandomGenerator failing at producing valid structure')
                return []
        
        return [candidate]
    
    def get_number_of_parents(self, sampler):
        return 0


"""
While generators generate candidates, postprocessors are used to apply common
postprocessing steps to all generated candidates. For example, the centering postprocessor
will center all candidates in the cell.

The most common postprocessors is the relax postprocessor, which performs a local
optimization on all candidates. Parallel implementations of this postprocessor are
also available and are recommended for searches that generated multiple candidates
pr. iteration.
"""
from .ABC_postprocess import PostprocessBaseClass
from .centering import CenteringPostProcess
from .disjoint_filtering import DisjointFilteringPostprocess
from .minimum_dist import MinimumDistPostProcess
from .ray_relax import ParallelRelaxPostprocess
from .relax import RelaxPostprocess
from .surface_centering import SurfaceCenteringPostprocess
from .wrap import WrapperPostprocess

__all__ = [
    'PostprocessBaseClass',
    'CenteringPostProcess',
    'DisjointFilteringPostprocess',
    'MinimumDistPostProcess',
    'ParallelRelaxPostprocess',
    'RelaxPostprocess',
    'SurfaceCenteringPostprocess',
    'WrapperPostprocess',
]



from agox.test.run_tests.run_utils import agox_test_run

def test_ea(tmp_path, cmd_options):
    mode = 'ea' # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options)


from typing import List

import numpy as np
from ase import Atoms

from agox.utils.filters.ABC_filter import FilterBaseClass


class NoneFilter(FilterBaseClass):
    """
    Filter that removes all structures.

    """

    name = "NoneFilter"

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        """
        super().__init__(**kwargs)

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        return np.array([])



"""
Models are used to predict the output of a given input. Generally this corresponds 
to predicting the energy and forces of a given atomic configuration. 

This module also contains the descriptors used to represent the atomic configurations.
"""
from .ABC_model import ModelBaseClass
from agox.models.GPR import GPR, SparseGPR, SparseGPREnsemble




from typing import Any, Dict, List, Tuple, Union

import numpy as np
from ase import Atoms
from ase.calculators.calculator import all_changes
from scipy.linalg import cho_factor, cho_solve
from scipy.optimize import fmin_l_bfgs_b

from agox.models.ABC_model import ModelBaseClass
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from agox.models.GPR.kernels import Kernel
from agox.utils import candidate_list_comprehension
from agox.utils.filters import EnergyFilter, FilterBaseClass
from agox.utils.ray import RayPoolUser

from timeit import default_timer as dt


class GPR(ModelBaseClass, RayPoolUser):

    """
    Gaussian Process Regression model without any approximations.

    Parameters
    ----------

    descriptor : DescriptorBaseClass
        Descriptor object, has to be a global descriptor.
    kernel : Kernel
        Kernel object to use.
    prior : ModelBaseClass
        Prior model object.
    single_atom_energies : dict
        Dictionary of single atom energies.
    n_optimize : int
        Number of LML optimization restarts.
    optimizer_maxiter : int
        Maximum number of iterations for the optimizer.
    centralize : bool
        Whether to centralize the energy.
    filter : FilterBaseClass
        Filter object to choose which structures to include in the training set.
    use_ray : bool
        Whether to use ray for parallelization of hyperparameter optimization.
    """

    name = "GPR"

    supported_descriptor_types = ["global"]
    implemented_properties = ["energy", "forces", "uncertainty", "forces_uncertainty"]

    def __init__(
        self,
        descriptor: DescriptorBaseClass,
        kernel: Kernel,
        prior: ModelBaseClass = None,
        single_atom_energies: Union[List[float], Dict[str, float]] = None,
        n_optimize: int = None,
        optimizer_maxiter: int = 100,
        centralize: bool = True,
        filter: FilterBaseClass = EnergyFilter(),
        use_ray=True,
        **kwargs
    ) -> None:
        """
        Parameters
        ----------
        descriptor : DescriptorBaseClass
            Descriptor object.
        kernel : KernelBaseClass
            Kernel object.
        feature_method : function
            Feature method.
        prior : ModelBaseClass
            Prior model object.
        single_atom_energies : dict
            Dictionary of single atom energies.
        centralize : bool
            Whether to centralize the energy.
        """
        ModelBaseClass.__init__(self, filter=filter, **kwargs)
        RayPoolUser.__init__(self)
        assert (
            descriptor.descriptor_type in self.supported_descriptor_types
        ), "Only global descriptors are supported, use SparseGPR for local descriptors instead."

        for attr in ["alpha", "K_inv", "X", "K", "kernel", "Y", "mean_energy", "_ready_state"]:
            self.add_dynamic_attribute(attr)

        self.descriptor = descriptor
        self.kernel = kernel
        self.prior = prior
        self.single_atom_energies = single_atom_energies

        self.n_optimize = n_optimize
        self.optimizer_maxiter = optimizer_maxiter
        self.centralize = centralize

        self.add_save_attributes(
            ["X", "Y", "mean_energy", "K", "K_inv", "alpha", "kernel.theta"]
        )

        # We add self to the pool, it will keep an updated copy of the model on the pool
        self.n_optimize = n_optimize
        self.use_ray = use_ray
        if self.use_ray:
            self.actor_model_key = self.pool_add_module(self)
            self.self_synchronizing = True  # Defaults to False, inherited from Module.
            if n_optimize is None:
                self.n_optimize = int(self.cpu_count)
        else:
            if n_optimize is None:
                self.n_optimize = 1

    def train(self, training_data: List[Atoms], **kwargs) -> None:
        """
        Train the model.

        Parameters
        ----------
        training_data : list
            List of Atoms objects.

        """
        
        self.training_timings = {}
        t_total = dt()

        self._data_before_filter = len(training_data)
        if self.filter is not None:
            training_data, _ = self.filter(training_data)
        self._data_after_filter = len(training_data)

        self.X, self.Y = self._preprocess(training_data)

        self._training_record(training_data)

        self._train_model(training_data)

        validation = self.validate()

        self.print_model_info(validation=validation)
        self.atoms = None

        self.training_timings["total"] = dt() - t_total

    @candidate_list_comprehension
    def predict_energy(
        self, atoms: Atoms, k: np.ndarray = None, x: np.ndarray = None, **kwargs
    ) -> float:
        if not self.ready_state:
            return self._postprocess_energy(atoms, 0)

        k = self._get_kernel_vector(k, x, atoms)

        e_pred = np.sum(k.T @ self.alpha)

        return self._postprocess_energy(atoms, e_pred)

    @candidate_list_comprehension
    def predict_uncertainty(
        self,
        atoms: Atoms,
        k: np.ndarray = None,
        k0: np.ndarray = None,
        x: np.ndarray = None,
        **kwargs
    ) -> float:
        """
        Predict uncertainty for a given atoms object

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object

        Returns
        -------
        float
            Uncertainty

        """
        if "uncertainty" not in self.implemented_properties or not self.ready_state:
            return 0

        if x is None:
            x = self._get_features(atoms)

        k = self._get_kernel_vector(k, x, atoms)

        if k0 is None:
            k0 = self.kernel(x, x)
        var = float(k0 - k.T @ self.K_inv @ k)
        return np.sqrt(max(var, 0))

    @candidate_list_comprehension
    def predict_forces(
        self, atoms: Atoms, dk_dr: np.ndarray = None, **kwargs
    ) -> np.ndarray:
        """
        Predict forces for a given atoms object

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object

        Returns
        -------
        np.ndarray
            Forces
        """
        if not self.ready_state:
            return self._postprocess_forces(atoms, np.zeros((len(atoms), 3)))

        if dk_dr is None:
            dk_dr = self._get_kernel_derivative(atoms)

        # Predict the forces:
        forces = (-dk_dr.sum(axis=0) @ self.alpha).squeeze()

        return self._postprocess_forces(atoms, forces)

    @candidate_list_comprehension
    def predict_uncertainty_forces(
        self,
        atoms: Atoms,
        x: np.ndarray = None,
        k: np.ndarray = None,
        k0: np.ndarray = None,
        dk_dr: np.ndarray = None,
        **kwargs
    ) -> np.ndarray:
        """
        Predict forces uncertainty for a given atoms object

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object

        Returns
        -------
        np.ndarray
            Forces uncertainty

        """

        if "forces_uncertainty" not in self.implemented_properties or not self.ready_state:
            return np.zeros((len(atoms), 3))

        if x is None:
            x = self._get_features(atoms)

        k = self._get_kernel_vector(k, x, atoms)

        if dk_dr is None:
            dk_dr = self._get_kernel_derivative(atoms).sum(axis=0)

        if k0 is None:
            k0 = self.kernel(x, x)

        var = k0 - k.T @ self.K_inv @ k
        if var < 0:
            std_force = np.zeros((len(atoms), 3))
        else:
            std_force = (1 / np.sqrt(var) * dk_dr @ self.K_inv @ k).squeeze()

        return std_force

    @candidate_list_comprehension
    def predict_energy_and_uncertainty(self, atoms, x=None, k=None, k0=None, **kwargs):
        
        if x is None:
            x = self._get_features(atoms)
            
        k = self._get_kernel_vector(k, x, atoms)

        energy = self.predict_energy(atoms, k=k, **kwargs)
        uncertainty = self.predict_uncertainty(atoms, k=k, k0=k0, x=x, **kwargs)

        return energy, uncertainty

    def converter(
        self, atoms: Atoms, derivatives=True, reduced: bool = False, **kwargs
    ) -> Dict[str, np.ndarray]:
        """
        Precompute all necessary quantities for the model

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object

        Returns
        -------
        dict
            Dictionary with all necessary quantities

        """
        x = self._get_features(atoms)
        k = self.kernel(self.X, x)
        k0 = self.kernel(x, x)

        d = {"x": x, "k": k, "k0": k0}
        if derivatives:
            dx_dr = self.descriptor.get_feature_gradient(atoms)
            dk_dr = self._get_kernel_derivative(atoms, x=x, dX_dr=dx_dr)
            d["dk_dr"] = dk_dr
            d["dx_dr"] = dx_dr

        return d

    @property
    def single_atom_energies(self) -> np.ndarray:
        """
        Get the single atom energies.

        Returns
        -------
        np.ndarray

        """
        return self._single_atom_energies

    @single_atom_energies.setter
    def single_atom_energies(self, s: Union[Dict, np.ndarray]) -> None:
        """
        Set the single atom energies.
        Index number corresponds to the atomic number ie. 1 = H, 2 = He, etc.

        Parameters
        ----------
        s : dict or np.ndarray
            Dictionary/array of single atom energies.

        """
        if isinstance(s, np.ndarray):
            self._single_atom_energies = s
        elif isinstance(s, dict):
            self._single_atom_energies = np.zeros(100)
            for i, val in s.items():
                self._single_atom_energies[i] = val
        elif s is None:
            self._single_atom_energies = np.zeros(100)

    def model_info(self, **kwargs) -> List[str]:
        """
        List of strings with model information
        """
        x = "    "
        filter_name = self.filter.name if self.filter is not None else "None"
        try:
            data_before_filter = self._data_before_filter
            data_after_filter = self._data_after_filter
            filter_removed_data = self._data_before_filter - self._data_after_filter
        except AttributeError:
            filter_removed_data = 0

        out = [
            "------ Model Info ------",
            "Descriptor:",
            x + "{}".format(self.descriptor.name),
            "Kernel:",
            x + "{}".format(self.kernel),
            "Filter:",
            x + "{} removed {} structures".format(filter_name, filter_removed_data),
            x + "- Data before filter: {}".format(data_before_filter),
            x + "- Data after filter: {}".format(data_after_filter),
            "------ Training Info ------",
            "Training data size: {}".format(self.X.shape[0]),
            "Neg. log marginal likelihood.: {:.2f}".format(self._nlml),
            "Hyperparameter optimizations: {}".format(self._nlml_opts),
        ]

        return out

    def calculate(
        self,
        atoms: Atoms = None,
        properties: List[str] = ["energy"],
        system_changes=all_changes,
    ) -> None:
        super().calculate(
            atoms=atoms, properties=properties, system_changes=system_changes
        )
        if "uncertainty" in properties:
            self.results["uncertainty"] = self.predict_uncertainty(atoms)
        if "force_uncertainty" in properties:
            self.results["force_uncertainty"] = self.predict_forces_uncertainty(atoms)

    def hyperparameter_search(self) -> None:
        """
        Hyperparameter search

        """
        initial_parameters = []
        initial_parameters.append(self.kernel.theta.copy())
        if self.n_optimize > 0:
            for _ in range(self.n_optimize - 1):
                init_theta = np.random.uniform(
                    size=(len(self.kernel.bounds),),
                    low=self.kernel.bounds[:, 0],
                    high=self.kernel.bounds[:, 1],
                )
                initial_parameters.append(init_theta)

            fmins = []
            thetas = []
            for init_theta in initial_parameters:
                theta_min, nll_min = self._hyperparameter_optimize(
                    init_theta=init_theta
                )
                fmins.append(nll_min)
                thetas.append(theta_min)

            self.kernel.theta = thetas[np.argmin(np.array(fmins))]
            self._nlml = np.min(fmins)
            self._nlml_opts = len(fmins)
        else:
            self._nlml = np.inf
            self._nlml_opts = 0

    def hyperparameter_search_parallel(self, relax=True) -> None:

        if self.n_optimize == 0:
            self._nlml = np.inf
            self._nlml_opts = 0
            return 

        bounds = self.kernel.bounds
        initial = np.random.uniform(size=(self.n_optimize, len(bounds)), low=bounds[:, 0], high=bounds[:, 1])
        initial[0, :] = self.kernel.theta

        modules = [
                [self.actor_model_key]
            ] * self.n_optimize  # All jobs use the same model that is already on the actor.

        # Create a list of all the arguments that should be passed to the function.
        args = [[init_theta] for init_theta in initial]
        kwargs = [{'relax':relax} for _ in range(self.n_optimize)]

        # Submit the jobs.
        outputs = self.pool_map(ray_hyperparameter_optimize, modules, args, kwargs)

        # Get the best theta
        likelihood = [output[1] for output in outputs]
        best_theta = outputs[np.argmin(likelihood)][0]

        self._nlml_opts = len(likelihood)

        # Set the best theta
        self.kernel.theta = best_theta
        self._nlml = np.min(likelihood)

    def monte_carlo_hyperparameter_search(self) -> None:
        """
        Hyperparameter search using Monte Carlo sampling

        """
        initial_parameters = []
        initial_parameters.append(self.kernel.theta.copy())
        if self.n_optimize > 0:
            for _ in range(self.n_optimize - 1):
                init_theta = np.random.uniform(
                    size=(len(self.kernel.bounds),),
                    low=self.kernel.bounds[:, 0],
                    high=self.kernel.bounds[:, 1],
                )
                initial_parameters.append(init_theta)

            fmins = []
            thetas = []
            for theta in initial_parameters:
                nlml = -self._log_marginal_likelihood(theta)
                fmins.append(nlml)
                thetas.append(theta)

            self.kernel.theta = thetas[np.argmin(np.array(fmins))]
            self._nlml = np.min(fmins)
        else:
            self._nlml = np.inf

    def _get_features(self, atoms: Atoms) -> np.ndarray:
        """
        Get features for a given atoms object

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object

        Returns
        -------
        np.ndarray
            Features

        """
        f = self.descriptor.get_features(atoms)     

        if isinstance(f, np.ndarray) and len(f.shape) == 1:
            f = f.reshape(1, -1)

        f = np.vstack(f)
        return f

    def _get_kernel_vector(self, k: np.ndarray = None, 
                        x: np.ndarray = None, 
                        atoms: Atoms = None) -> np.ndarray:
        """
        Get the kernel vector for a given atoms object

        Parameters
        ----------
        k : np.ndarray, optional
            Kernel vector, by default None
        x : np.ndarray, optional
            Features, by default None
        atoms : ase.Atoms, optional
            Atoms object, by default None

        Returns
        -------
        np.ndarray
            Kernel vector
        """
        
        if k is not None:
            return k

        if x is None:
            x = self._get_features(atoms)
        
        return self.kernel(self.X, x)

    def _get_kernel_derivative(self, atoms: Atoms, 
            x: np.ndarray = None, 
            dX_dr: np.ndarray = None, 
            dk_dX: np.ndarray = None) -> np.ndarray:
        """
        Get the kernel derivative wrt. the Carteisian coordinates r.

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object
        x : np.ndarray, optional
            Features, by default None
        dX_dr : np.ndarray, optional
            Derivatives of the features wrt. the Carteisian coordinates r, by default None
        dk_dX : np.ndarray, optional
            Derivatives of the kernel wrt. the features, by default None
        
        Returns
        -------
        np.ndarray
            Derivatives of the kernel wrt. the Carteisian coordinates r.
        """
        # Get features and derivatives:
        if x is None:
            x = self._get_features(atoms)  # [centers, features]

        if dX_dr is None:
            dX_dr = self.descriptor.get_feature_gradient(
                atoms
            )  # [centers, atoms, 3, features]

        # Get kernel derivatives:
        if dk_dX is None:
            dk_dX = np.array(
                [
                    self.kernel.get_feature_gradient(self.X, x[i : i + 1])
                    for i in range(len(x))
                ]
            )

        # Convert to derivatives wrt. Carteisian coordinates r.
        dk_dr = np.einsum(
            "nadf,nmf->nadm", dX_dr, dk_dX
        )  # [centers, atoms, 3, features]

        return dk_dr

    def _train_model(self, training_data: List[Atoms], synchronize=True) -> None:
        """
        Train the model

        """
        if self.use_ray:
            self.pool_synchronize(attributes=["X", "Y"], writer=self.writer)
            self.hyperparameter_search_parallel()
        else:
            self.hyperparameter_search()

        self.K = self.kernel(self.X)
        self.alpha, self.K_inv, _ = self._solve(self.K, self.Y)
        self.ready_state = True

        if self.use_ray and synchronize:
            self.pool_synchronize(
                attributes=["alpha", "K_inv", "kernel", "K", "mean_energy", "_ready_state"], writer=self.writer
            )

    def _preprocess(self, data: List[Atoms]) -> None:
        """
        Preprocess the training data.

        Parameters
        ----------
        data : list
            List of Atoms objects.

        Returns
        -------
        np.ndarray
            Features.
        np.ndarray
            Targets.

        """
        self._N_energies = len(data)
        Y = np.expand_dims(
            np.array(
                [
                    d.get_potential_energy()
                    - sum(self.single_atom_energies[d.get_atomic_numbers()])
                    for d in data
                ]
            ),
            axis=1,
        )

        # Calculate the prior:
        if self.prior is None:
            self.prior_energy = np.zeros(Y.shape)
        else:
            self.prior_energy = np.expand_dims(
                np.array([self.prior.predict_energy(d) for d in data]), axis=1
            )

        Y -= self.prior_energy

        # Energy centralization:
        if self.centralize:
            self.mean_energy = np.mean(Y)
        else:
            self.mean_energy = 0

        Y -= self.mean_energy

        # Calculate features:
        t_feature = dt()
        X = self._get_features(data)
        self.training_timings["feature"] = dt() - t_feature

        return X, Y

    def _postprocess_energy(self, atoms: Atoms, e_pred: float) -> float:
        """
        Postprocess the energy.

        Parameters
        ----------
        atoms : Atoms
            Atoms object.
        e_pred : float
            Predicted energy.

        Returns
        -------
        float
            Postprocessed energy.
        """
        if self.prior is None:
            prior = 0
        else:
            prior = self.prior.predict_energy(atoms)

        sae = sum(self.single_atom_energies[atoms.get_atomic_numbers()])

        total = e_pred + prior + sae

        if hasattr(self, "mean_energy"):
            total += self.mean_energy

        return total

    def _postprocess_forces(self, atoms: Atoms, f_pred: np.ndarray) -> np.ndarray:
        """
        Postprocess the forces.

        Parameters
        ----------
        atoms : Atoms
            Atoms object.
        f_pred : np.ndarray
            Predicted forces.

        Returns
        -------
        np.ndarray
            Postprocessed forces.
        """
        if self.prior is None:
            prior = 0
        else:
            prior = self.prior.predict_forces(atoms)

        return f_pred + prior

    def _hyperparameter_optimize(
        self, init_theta: np.ndarray = None
    ) -> Tuple[np.ndarray, float]:
        """
        Hyperparameter optimization

        Parameters
        ----------
        init_theta : np.ndarray
            Initial theta

        Returns
        -------
        np.ndarray
            Optimal theta

        """

        def f(theta):
            P, grad_P = self._log_marginal_likelihood_gradient(theta)
            if np.isnan(P):
                return np.inf, np.zeros_like(theta, dtype="float64")
            P, grad_P = -float(P), -np.asarray(grad_P, dtype="float64")
            return P, grad_P

        bounds = self.kernel.bounds

        if init_theta is None:
            self.key, key = np.random.split(self.key)
            init_theta = np.random.uniform(
                key, shape=(len(bounds),), minval=bounds[:, 0], maxval=bounds[:, 1]
            )

        theta_min, fmin, conv = fmin_l_bfgs_b(
            f,
            np.asarray(init_theta, dtype="float64"),
            bounds=np.asarray(bounds, dtype="float64"),
            maxiter=self.optimizer_maxiter,
        )

        return theta_min, fmin

    def _solve(
        self, K: np.ndarray, Y: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Solve the linear system

        Parameters
        ----------
        K : np.ndarray
            Kernel matrix
        Y : np.ndarray
            Target values

        Returns
        -------
        np.ndarray
            Alpha
        np.ndarray
            Inverse of kernel matrix
        np.ndarray
            Cholesky decomposition of kernel matrix

        """
        L, lower = cho_factor(K)
        alpha = cho_solve((L, lower), Y)
        K_inv = cho_solve((L, lower), np.eye(K.shape[0]))
        return alpha, K_inv, (L, lower)

    def _log_marginal_likelihood(self, theta: np.ndarray) -> float:
        """
        Marginal log likelihood

        Parameters
        ----------
        theta : np.ndarray
            Kernel parameters

        Returns
        -------
        float
            log Marginal likelihood

        """
        t = self.kernel.theta.copy()
        self.kernel.theta = theta
        K = self.kernel(self.X)
        self.kernel.theta = t

        alpha, K_inv, (L, lower) = self._solve(K, self.Y)
        

        log_P = (
            -0.5 * np.einsum("ik,ik->k", self.Y, alpha)
            - np.sum(np.log(np.diag(L)))
            - K.shape[0] / 2 * np.log(2 * np.pi)
        )

        return np.sum(log_P)

    def _log_marginal_likelihood_gradient(
        self, theta: np.ndarray
    ) -> Tuple[float, np.ndarray]:
        """
        Marginal log likelihood gradient

        Parameters
        ----------
        theta : np.ndarray
            Kernel parameters

        Returns
        -------
        float
            Marginal log likelihood
        np.ndarray
            Marginal log likelihood gradient

        """
        t = self.kernel.theta.copy()
        self.kernel.theta = theta
        K, K_hp_gradient = self.kernel(self.X, eval_gradient=True)
        self.kernel.theta = t

        alpha, K_inv, (L, lower) = self._solve(K, self.Y)

        log_P = (
            -0.5 * np.einsum("ik,ik->k", self.Y, alpha)
            - np.sum(np.log(np.diag(L)))
            - K.shape[0] / 2 * np.log(2 * np.pi)
        )

        inner = np.squeeze(np.einsum("ik,jk->ijk", alpha, alpha), axis=2) - K_inv
        inner = inner[:, :, np.newaxis]

        grad_log_P = np.sum(
            0.5 * np.einsum("ijl,ijk->kl", inner, K_hp_gradient), axis=-1
        )
        return log_P, grad_log_P


def ray_hyperparameter_optimize(model, init_theta, relax=True):
    """
    Hyperparameter optimization

    Parameters
    ----------
    model : Model
        GPR model
    n_opt : int
        Number of optimization steps
    use_current_theta : bool
        Use current theta as initial theta

    Returns
    -------
    np.ndarray
        Optimal theta
    float
        Optimal log marginal likelihoodg

    """
    
    bounds = model.kernel.bounds

    def f(theta):
        P, grad_P = model._log_marginal_likelihood_gradient(theta)
        if np.isnan(P):
            return np.inf, np.zeros_like(theta, dtype="float64")
        P, grad_P = -float(P), -np.asarray(grad_P, dtype="float64")
        return P, grad_P

    if relax:
        theta_min, fmin, conv = fmin_l_bfgs_b(
            f,
            np.asarray(init_theta, dtype="float64"),
            bounds=np.asarray(bounds, dtype="float64"),
            maxiter=model.optimizer_maxiter,
        )
    else:
        theta_min = init_theta
        fmin = -model._log_marginal_likelihood(init_theta)

    return theta_min, fmin



import pytest
import numpy as np
from agox.candidates import CandidateBaseClass
from agox.helpers.confinement import Confinement
from agox.generators import SymmetryGenerator, SymmetryPermutationGenerator, SymmetryRattleGenerator

@pytest.fixture(scope='module')
def random_symmetry_cluster_generator():

    def make_generator(environment):
        generator = SymmetryGenerator(**environment.get_confinement(), sym_type='cluster')
        return generator
    return make_generator

@pytest.fixture(scope='module')
def random_symmetry_cluster_candidate(random_symmetry_cluster_generator, test_environment_non_periodic):
    generator = random_symmetry_cluster_generator(test_environment_non_periodic)

    candidate = None
    attempt = 0 
    while candidate is None and attempt < 100: # This is bad, but...
        generator_output = generator(None, test_environment_non_periodic)
        attempt += 1

        if len(generator_output) > 0:
            candidate = generator_output[0]

    return candidate        

def test_output_type(random_symmetry_cluster_candidate):
    assert isinstance(random_symmetry_cluster_candidate, CandidateBaseClass)

def test_output_symmetry_type(random_symmetry_cluster_candidate):
    space_group = random_symmetry_cluster_candidate.get_meta_information('space_group')
    assert isinstance(space_group, str)

def test_confinement(random_symmetry_cluster_candidate, test_environment_non_periodic):
    indices = test_environment_non_periodic.get_missing_indices()
    positions = random_symmetry_cluster_candidate.positions[indices, :]
    confinement = Confinement(**test_environment_non_periodic.get_confinement())
    assert confinement.check_confinement(positions).all()

def test_template(random_symmetry_cluster_candidate, test_environment_non_periodic):
    template = test_environment_non_periodic.get_template()
    assert (random_symmetry_cluster_candidate.positions[0:len(template)] == template.positions).all()

def test_cell(random_symmetry_cluster_candidate, test_environment_non_periodic):
    assert (random_symmetry_cluster_candidate.cell == test_environment_non_periodic.get_template().get_cell()).all()



from typing import Tuple
from statistics import NormalDist
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats

from scipy.stats import linregress


def calculate_density(
    norm: stats.norm, stdevs: np.ndarray, residuals: np.ndarray, percentile: float
) -> float:
    """
    Calculate the fraction of the residuals that fall within the lower
    `percentile` of their respective Gaussian distributions, which are
    defined by their respective uncertainty estimates.

    Parameters
    ----------
    norm : scipy.stats.norm
        The normal distribution to use for the calculation.
    stdevs : np.ndarray
        The standard deviations of the residuals.
    residuals : np.ndarray
        The residuals to calculate the density of.
    percentile : float
        The percentile to calculate the density of.
    """
    # Find the normalized bounds of this percentile
    upper_bound = norm.ppf(percentile)

    # Normalize the residuals so they all should fall on the normal bell curve
    normalized_residuals = residuals.reshape(-1) / stdevs.reshape(-1)

    # Count how many residuals fall inside here
    num_within_quantile = 0
    for resid in normalized_residuals:
        if resid <= upper_bound:
            num_within_quantile += 1

    # Return the fraction of residuals that fall within the bounds
    density = num_within_quantile / len(residuals)
    return density


def calibration_curve(true, pred, stdevs, bins=100) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the calibration curve of a set of predictions.

    Parameters
    ----------
    true : np.ndarray
        The true values.
    pred : np.ndarray
        The predicted values.
    stdevs : np.ndarray
        The standard deviations of the predictions.
    bins : int
        The number of bins to use for the calibration curve.
    """

    residuals = true - pred
    norm = stats.norm(0, 1)
    predicted_pi = np.linspace(0, 1, bins)
    observed_pi = [
        calculate_density(norm, stdevs, residuals, percentile)
        for percentile in predicted_pi
    ]

    return predicted_pi, observed_pi


def miscalibration_area(predicted_pi: np.ndarray, observed_pi: np.ndarray) -> float:
    """
    Calculate the area between the calibration curve and the line of perfect calibration.

    Parameters
    ----------
    predicted_pi : np.ndarray
        The predicted probabilities.
    observed_pi : np.ndarray
        The observed probabilities.
    """
    return np.trapz(np.abs(predicted_pi - observed_pi), predicted_pi)


def calibration_error(predicted_pi: np.ndarray, observed_pi: np.ndarray) -> float:
    """
    Calculate the calibration error of a calibration curve.

    Parameters
    ----------
    predicted_pi : np.ndarray
        The predicted probabilities.
    observed_pi : np.ndarray
        The observed probabilities.
    """
    return ((predicted_pi - observed_pi) ** 2).sum()


def sharpness(stdevs: np.ndarray):
    """
    Calculate the sharpness of the uncertainty estimates.
    Here we calculate it as the square-root of the mean of the squared to have
    it in the same units as the predictions.
    """
    return np.sqrt((stdevs**2).mean())


def dispersion(stdevs: np.ndarray):
    """
    Calculate the dispersion of the uncertainty estimates.
    """
    return (
        np.sqrt(((stdevs - stdevs.mean()) ** 2).sum() / (len(stdevs) - 1))
        / stdevs.mean()
    )


def error_based_calibration(
    truth: np.ndarray,
    pred: np.ndarray,
    sigma: np.ndarray,
    bins: int = 10,
    sigma_max: float = None,
    sigma_min: float = None,
):
    """
    Calculate the error-based calibration of a set of predictions.

    Parameters
    ----------
    truth : np.ndarray
        The true values.
    pred : np.ndarray
        The predicted values.
    sigma : np.ndarray
        The standard deviations of the predictions.
    bins : int
        The number of bins to use.
    sigma_max : float
        The maximum standard deviation to consider.
    sigma_min : float
        The minimum standard deviation.
    """

    eps = truth - pred
    if sigma_max is None:
        sigma_max = sigma.max()
    if sigma_min is None:
        sigma_min = sigma.min()

    bin_edges = np.linspace(sigma_min, sigma_max, bins)

    rmse = []
    rmv = []

    for i in range(bins - 1):
        mask = (sigma > bin_edges[i]) * (sigma <= bin_edges[i + 1])
        count = np.sum(mask)

        if count == 0:
            continue

        rmse.append(np.sqrt(1 / count * np.sum(eps[mask] ** 2)))
        rmv.append(np.sqrt(1 / count * np.sum(sigma[mask] ** 2)))

    rmse = np.array(rmse)
    rmv = np.array(rmv)

    return rmse, rmv


def calibration_plot(true, pred, stdevs, ax=None, **kwargs):
    predicted_pi, observed_pi = calibration_curve(true, pred, stdevs)

    if ax is None:
        fig, ax = plt.subplots()

    (l1,) = ax.plot(predicted_pi, observed_pi, **kwargs)
    ax.plot([0, 1], [0, 1], "k--", alpha=0.5)

    # Fill between:
    ax.fill_between(
        predicted_pi, predicted_pi, observed_pi, alpha=0.99, color=l1.get_color()
    )

    mis_area = miscalibration_area(predicted_pi, observed_pi)
    cal_error = calibration_error(predicted_pi, observed_pi)

    # text = f'Miscalibration area: {mis_area:.3f}\nCalibration error: {cal_error:.3f}'
    # ax.text(0.05, 0.95, text, transform=ax.transAxes, va='top', ha='left')

    ax.set_xlabel("Predicted probability")
    ax.set_ylabel("Observed probability")
    ax.set_xlim([0, 1])

    return ax


def plot_residuals(ax, true, pred, sigmas, sigma_bin=1, steps=10):
    residuals = pred - true

    l1 = ax.scatter(residuals, sigmas)
    s = np.linspace(0, 20)
    ax.set_xlim([-20, 20])
    ax.set_ylim([0, 20])

    for conf in [1.96]:
        for sign in [1, -1]:
            ax.plot(
                sign * conf * s,
                s,
                color="black",
                linestyle="--",
                label="95% confidence interval",
            )
    ax.set_xlabel("Residual")
    ax.set_ylabel("Sigma")

    for i in range(1, steps):
        indices = (sigmas < sigma_bin * i) & (sigmas > sigma_bin * (i - 1))
        res = residuals[indices]

        norm = NormalDist.from_samples(res)

        weights, bins = np.histogram(res, bins=50)
        weights = weights / np.max(weights) * sigma_bin
        ax.bar(
            bins[:-1],
            weights,
            width=bins[1] - bins[0],
            alpha=0.9,
            bottom=sigma_bin * (i - 1),
            color="red",
            edgecolor="black",
        )
        x = np.linspace(-norm.stdev * 1.96, norm.stdev * 1.96, 100)
        y = sigma_bin * np.exp(-(x**2) / (2 * norm.stdev**2)) + sigma_bin * (i - 1)
        ax.plot(x, y, color="red", linestyle="--")


def plot_error_based_calibration(
    ax, truth, prediction, sigma, bins=10, sigma_min=None, sigma_max=None
):
    rmse, rmv = error_based_calibration(
        truth, prediction, sigma, bins=bins, sigma_min=sigma_min, sigma_max=sigma_max
    )

    ax.plot(rmv, rmse, "-o")

    lin_fit = linregress(rmv, rmse)

    ax.plot(
        rmv,
        lin_fit.intercept + lin_fit.slope * rmv,
        "--",
        color="black",
        label=f"a = {lin_fit.slope:.2f}, b = {lin_fit.intercept:.2f}",
    )
    ax.plot(rmv, 0 + 1 * rmv, "r--", label="Ideal")

    ax.legend()
    ax.set_xlabel("RMV [eV]")
    ax.set_ylabel("RMSE [eV]")


def sharpness_plot(stdevs, bins=25, ax=None):
    if ax is None:
        fig, ax = plt.subplots()

    ax.hist(stdevs, bins=bins, density=True, edgecolor="k")

    ax.set_xlabel("$\sigma$")
    ax.set_ylabel("Density")

    sharpness_val = sharpness(stdevs)
    ax.axvline(sharpness_val, color="k", linestyle="--")

    dispersion_val = dispersion(stdevs)

    text = f"Sharpness: {sharpness_val:.3f}\nDispersion: {dispersion_val:.3f}"
    ax.text(0.95, 0.95, text, transform=ax.transAxes, va="top", ha="right")
    return ax



import pytest
from agox.test.run_tests.run_utils import agox_test_run

@pytest.mark.ray
def test_lgpr_bh(tmp_path, cmd_options):
    # pytest.skip('This test is not yet working always due to tolerances on different machines.')
    mode = 'lgpr_bh' # This determines the script that is imported.    
    options = {'tolerance':{'rtol':1e-03, 'atol':1e-03}, 
               'create_mode':cmd_options['create_mode']}
    agox_test_run(mode, tmp_path, options)
